# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
"""
Schur symmetric functions
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#                     2012 Mike Zabrocki <mike.zabrocki@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.arith.misc import factorial
from sage.combinat.tableau import StandardTableaux
from sage.data_structures.blas_dict import convert_remove_zeroes
from sage.misc.lazy_import import lazy_import
from sage.misc.misc_c import prod
from sage.rings.infinity import infinity
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

from . import classical

lazy_import('sage.libs.lrcalc', 'lrcalc')


class SymmetricFunctionAlgebra_schur(classical.SymmetricFunctionAlgebra_classical):
    def __init__(self, Sym):
        """
        A class for methods related to the Schur symmetric function basis.

        INPUT:

        - ``self`` -- a Schur symmetric function basis
        - ``Sym`` -- an instance of the ring of the symmetric functions

        TESTS::

            sage: s = SymmetricFunctions(QQ).s()
            sage: s == loads(dumps(s))
            True
            sage: TestSuite(s).run(skip=['_test_associativity', '_test_distributivity', '_test_prod'])  # needs lrcalc_python
            sage: TestSuite(s).run(elements=[s[1,1]+s[2], s[1]+2*s[1,1]])                               # needs lrcalc_python
        """
        classical.SymmetricFunctionAlgebra_classical.__init__(self, Sym, "Schur", 's')

    def _dual_basis_default(self):
        """
        Return the default value for ``self.dual_basis()``.

        This method returns the dual basis to the Schur basis with respect to the standard
        scalar product. Since the Schur basis is self-dual, it returns itself.

        EXAMPLES::

            sage: s = SymmetricFunctions(QQ).s()
            sage: ds = s.dual_basis()
            sage: s is ds
            True

            sage: zee = lambda x : x.centralizer_size()
            sage: S = s.dual_basis(zee); S
            Dual basis to Symmetric Functions over Rational Field in the Schur basis
            sage: S[2,1].scalar(s[2,1])
            1

        TESTS::

            sage: s._dual_basis_default() is s.dual_basis()
            True
        """
        return self

    def product_on_basis(self, left, right):
        """
        Return the product of ``left`` and ``right``.

        INPUT:

        - ``self`` -- a Schur symmetric function basis
        - ``left``, ``right`` -- partitions

        OUTPUT: an element of the Schur basis, the product of ``left`` and ``right``

        TESTS::

            sage: # needs lrcalc_python
            sage: s = SymmetricFunctions(QQ).s()
            sage: a = s([2,1]) + 1; a
            s[] + s[2, 1]
            sage: a^2   # indirect doctest
            s[] + 2*s[2, 1] + s[2, 2, 1, 1] + s[2, 2, 2] + s[3, 1, 1, 1]
             + 2*s[3, 2, 1] + s[3, 3] + s[4, 1, 1] + s[4, 2]

        Examples failing with three different messages in symmetrica::

            sage: # needs lrcalc_python
            sage: s[123,1]*s[1,1]
            s[123, 1, 1, 1] + s[123, 2, 1] + s[124, 1, 1] + s[124, 2]
            sage: s[123]*s[2,1]
            s[123, 2, 1] + s[124, 1, 1] + s[124, 2] + s[125, 1]
            sage: s[125]*s[3]
            s[125, 3] + s[126, 2] + s[127, 1] + s[128]

        ::

            sage: # needs lrcalc_python
            sage: QQx.<x> = QQ[]
            sage: s = SymmetricFunctions(QQx).s()
            sage: a = x^2*s([2,1]) + 2*x; a
            2*x*s[] + x^2*s[2, 1]
            sage: a^2
            4*x^2*s[] + 4*x^3*s[2, 1] + x^4*s[2, 2, 1, 1] + x^4*s[2, 2, 2]
             + x^4*s[3, 1, 1, 1] + 2*x^4*s[3, 2, 1] + x^4*s[3, 3]
             + x^4*s[4, 1, 1] + x^4*s[4, 2]

        ::

            sage: 0*s([2,1])                                                            # needs lrcalc_python
            0

        Example over a field with positive characteristic::

            sage: # needs lrcalc_python
            sage: s[2,1]^2
            s[2, 2, 1, 1] + s[2, 2, 2] + s[3, 1, 1, 1] + 2*s[3, 2, 1]
             + s[3, 3] + s[4, 1, 1] + s[4, 2]
            sage: s = SymmetricFunctions(GF(2)).s()
            sage: s[2,1]^2
            s[2, 2, 1, 1] + s[2, 2, 2] + s[3, 1, 1, 1] + s[3, 3] + s[4, 1, 1] + s[4, 2]
        """
        return self.element_class(self, convert_remove_zeroes(lrcalc.mult(left, right),
                                                            self.base_ring()))

    def coproduct_on_basis(self, mu):
        r"""
        Return the coproduct of ``self(mu)``.

        Here ``self`` is the basis of Schur functions in the ring of symmetric functions.

        INPUT:

        - ``self`` -- a Schur symmetric function basis
        - ``mu`` -- a partition

        OUTPUT:

        - the image of the ``mu``-th Schur function under the comultiplication of
          the Hopf algebra of symmetric functions; this is an element of the
          tensor square of the Schur basis

        EXAMPLES::

            sage: # needs lrcalc_python
            sage: Sym = SymmetricFunctions(QQ)
            sage: s = Sym.schur()
            sage: s.coproduct_on_basis([2])
            s[] # s[2] + s[1] # s[1] + s[2] # s[]

        TESTS::

            sage: s = SymmetricFunctions(QQ['t']).s()                                   # needs lrcalc_python
            sage: s[2].coproduct() / 2                                                  # needs lrcalc_python
            1/2*s[] # s[2] + 1/2*s[1] # s[1] + 1/2*s[2] # s[]
        """
        T = self.tensor_square()
        return T.element_class(T, convert_remove_zeroes(lrcalc.coprod(mu, all=1),
                                                      self.base_ring()))

    def _element_constructor_(self, x):
        """
        Construct an element of ``self`` from ``x``.

        TESTS::

            sage: s = SymmetricFunctions(QQ).s()
            sage: s([[2,1],[1]])                                                        # needs lrcalc_python
            s[1, 1] + s[2]
            sage: s([[],[]])                                                            # needs lrcalc_python
            s[]
        """
        ###################
        # Skew Partitions #
        ###################
        try:
            return self.skew_schur(x)
        except ValueError:
            return super()._element_constructor_(x)

    def _repeated_bernstein_creation_operator_on_basis(self, la, nu):
        r"""
        A Schur function indexed by a partition or zero from applying creation
        operators on `s(la)`.

        INPUT:

        - ``la`` -- a partition
        - ``nu`` -- list of lintegers

        EXAMPLES::

            sage: s = SymmetricFunctions(QQ).schur()
            sage: rbco = s._repeated_bernstein_creation_operator_on_basis
            sage: rbco(Partition([2,1]),[1])
            0
            sage: rbco(Partition([2,1]),[2])
            s[2, 2, 1]
            sage: rbco(Partition([2,1]),[-2])
            s[1]
            sage: rbco(Partition([2,1]),[1, -2])
            s[1, 1]
            sage: rbco(Partition([2,1]),[1, 0])
            -s[1, 1, 1, 1]
            sage: rbco(Partition([2,1]),[-3, 0])
            s[]
        """
        r = len(nu) + len(la)
        ga = [a-b for (a,b) in zip(nu+la.to_list(), range(-r,0))]
        if r == len(set(ga)) and min(ga) > 0:
            m = sum(1 for i in range(len(ga)) for j in range(i, len(ga))
                    if ga[i] < ga[j])
            ga.sort(reverse=True)
            return (-1)**m * self([a+b for (a,b) in zip(ga, range(-r,0))])
        return self.zero()

    def _magma_init_(self, magma):
        """
        Used in converting this ring to the corresponding ring in MAGMA.

        EXAMPLES::

            sage: # optional - magma
            sage: S = SymmetricFunctions(QQ).s()
            sage: t = 4*S[3,2]+9
            sage: mt = magma(t); mt
            9 + 4*$.[3,2]
            sage: mt.sage()
            9*s[] + 4*s[3, 2]
        """
        B = magma(self.base_ring())
        Bref = B._ref()
        return f"SymmetricFunctionAlgebraSchur({Bref})"

    class Element(classical.SymmetricFunctionAlgebra_classical.Element):
        def __pow__(self, n):
            """
            Return the naive powering of an instance of ``self``.

            INPUT:

            - ``self`` -- an element of the Schur symmetric function basis
            - ``n`` -- nonnegative integer

            OUTPUT: the `n`-th power of an instance of ``self`` in the Schur basis

            See ``Monoids.Element.__pow__`` and ``Monoids.Element._pow_naive``.

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ['x']).s()
                sage: len(s([2,1])^8) # long time (~ 4 s)
                1485
                sage: len(s([2,1])^9) # long time (~10 s)
                2876

            Binary exponentiation does not seem to bring any speedup for
            Schur functions. This most likely is because of the
            explosion of the number of terms.

            #    sage: s = SymmetricFunctions(QQ).s(); y = s([1])
            #    sage: n = 24
            #    sage: %timeit y**n    # using binary exponentiation
            #    10 loops, best of 3: 1.22 s per loop
            #    sage: %timeit prod(y for i in range(n))
            #    10 loops, best of 3: 1.06 s per loop

            With polynomial coefficients, this is actually much *slower*
            (although this should be profiled further; there seems to
            be an unreasonable number of polynomial multiplication involved,
            besides the fact that 1 * QQ['x'].one() currently involves a
            polynomial multiplication)

            #    sage: sage: s = SymmetricFunctions(QQ['x']).s()
            #    sage: y = s([2,1])
            #    sage: %timeit y**7
            #    10 loops, best of 3: 18.9 s per loop
            #    sage: %timeit y*y*y*y*y*y*y
            #    10 loops, best of 3: 1.73 s per loop

            Todo: do the same for the other non multiplicative bases?
            """
            return self._pow_naive(n)

        def omega(self):
            r"""
            Return the image of ``self`` under the omega automorphism.

            The *omega automorphism* is defined to be the unique algebra
            endomorphism `\omega` of the ring of symmetric functions that
            satisfies `\omega(e_k) = h_k` for all positive integers `k`
            (where `e_k` stands for the `k`-th elementary symmetric
            function, and `h_k` stands for the `k`-th complete homogeneous
            symmetric function). It furthermore is a Hopf algebra
            endomorphism and an involution, and it is also known as the
            *omega involution*. It sends the power-sum symmetric function
            `p_k` to `(-1)^{k-1} p_k` for every positive integer `k`.

            The images of some bases under the omega automorphism are given by

            .. MATH::

                \omega(e_{\lambda}) = h_{\lambda}, \qquad
                \omega(h_{\lambda}) = e_{\lambda}, \qquad
                \omega(p_{\lambda}) = (-1)^{|\lambda| - \ell(\lambda)}
                p_{\lambda}, \qquad
                \omega(s_{\lambda}) = s_{\lambda^{\prime}},

            where `\lambda` is any partition, where `\ell(\lambda)` denotes
            the length (:meth:`~sage.combinat.partition.Partition.length`)
            of the partition `\lambda`, where `\lambda^{\prime}` denotes the
            conjugate partition
            (:meth:`~sage.combinat.partition.Partition.conjugate`) of
            `\lambda`, and where the usual notations for bases are used
            (`e` = elementary, `h` = complete homogeneous, `p` = powersum,
            `s` = Schur).

            :meth:`omega_involution()` is a synonym for the :meth:`omega()`
            method.

            OUTPUT: the image of ``self`` under the omega automorphism

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ).s()
                sage: s([2,1]).omega()
                s[2, 1]
                sage: s([2,1,1]).omega()
                s[3, 1]
            """
            conj = lambda part: part.conjugate()
            return self.map_support(conj)

        omega_involution = omega

        def scalar(self, x, zee=None):
            r"""
            Return the standard scalar product between ``self`` and `x`.

            Note that the Schur functions are self-dual with respect to this
            scalar product. They are also lower-triangularly related to the
            monomial symmetric functions with respect to this scalar product.

            INPUT:

            - ``x`` -- element of the ring of symmetric functions over the
              same base ring as ``self``

            - ``zee`` -- an optional function on partitions giving
              the value for the scalar product between the power-sum
              symmetric function `p_{\mu}` and itself
              (the default value is the standard
              :meth:`~sage.combinat.sf.sfa.zee` function)

            OUTPUT: the scalar product between ``self`` and ``x``

            EXAMPLES::

                sage: s = SymmetricFunctions(ZZ).s()
                sage: a = s([2,1])
                sage: b = s([1,1,1])
                sage: c = 2*s([1,1,1])
                sage: d = a + b
                sage: a.scalar(a)
                1
                sage: b.scalar(b)
                1
                sage: b.scalar(a)
                0
                sage: b.scalar(c)
                2
                sage: c.scalar(c)
                4
                sage: d.scalar(a)
                1
                sage: d.scalar(b)
                1
                sage: d.scalar(c)
                2

            ::

                sage: m = SymmetricFunctions(ZZ).monomial()
                sage: p4 = Partitions(4)
                sage: l = [ [s(p).scalar(m(q)) for q in p4] for p in p4]
                sage: matrix(l)
                [ 1  0  0  0  0]
                [-1  1  0  0  0]
                [ 0 -1  1  0  0]
                [ 1 -1 -1  1  0]
                [-1  2  1 -3  1]
            """
            if zee is None:
                s = self.parent()
                R = s.base_ring()
                one = R.one()
                f = lambda p1, p2: one
                x = s(x)
                return s._apply_multi_module_morphism(self, x, f, orthogonal=True)
            else:
                p = self.parent().realization_of().power()
                return p(self).scalar( x, zee=zee )

        def verschiebung(self, n):
            r"""
            Return the image of the symmetric function ``self`` under the
            `n`-th Verschiebung operator.

            The `n`-th Verschiebung operator `\mathbf{V}_n` is defined to be
            the unique algebra endomorphism `V` of the ring of symmetric
            functions that satisfies `V(h_r) = h_{r/n}` for every positive
            integer `r` divisible by `n`, and satisfies `V(h_r) = 0` for
            every positive integer `r` not divisible by `n`. This operator
            `\mathbf{V}_n` is a Hopf algebra endomorphism. For every
            nonnegative integer `r` with `n \mid r`, it satisfies

            .. MATH::

                \mathbf{V}_n(h_r) = h_{r/n},
                \quad \mathbf{V}_n(p_r) = n p_{r/n},
                \quad \mathbf{V}_n(e_r) = (-1)^{r - r/n} e_{r/n}

            (where `h` is the complete homogeneous basis, `p` is the
            powersum basis, and `e` is the elementary basis). For every
            nonnegative integer `r` with `n \nmid r`, it satisfes

            .. MATH::

                \mathbf{V}_n(h_r) = \mathbf{V}_n(p_r) = \mathbf{V}_n(e_r) = 0.

            The `n`-th Verschiebung operator is also called the `n`-th
            Verschiebung endomorphism. Its name derives from the Verschiebung
            (German for "shift") endomorphism of the Witt vectors.

            The `n`-th Verschiebung operator is adjoint to the `n`-th
            Frobenius operator (see :meth:`frobenius` for its definition)
            with respect to the Hall scalar product (:meth:`scalar`).

            The action of the `n`-th Verschiebung operator on the Schur basis
            can also be computed explicitly. The following (probably clumsier
            than necessary) description can be obtained by solving exercise
            7.61 in Stanley's [STA]_.

            Let `\lambda` be a partition. Let `n` be a positive integer. If
            the `n`-core of `\lambda` is nonempty, then
            `\mathbf{V}_n(s_\lambda) = 0`. Otherwise, the following method
            computes `\mathbf{V}_n(s_\lambda)`: Write the partition `\lambda`
            in the form `(\lambda_1, \lambda_2, \ldots, \lambda_{ns})` for some
            nonnegative integer `s`. (If `n` does not divide the length of
            `\lambda`, then this is achieved by adding trailing zeroes to
            `\lambda`.) Set `\beta_i = \lambda_i + ns - i` for every
            `s \in \{ 1, 2, \ldots, ns \}`. Then,
            `(\beta_1, \beta_2, \ldots, \beta_{ns})` is a strictly decreasing
            sequence of nonnegative integers. Stably sort the list
            `(1, 2, \ldots, ns)` in order of (weakly) increasing remainder of
            `-1 - \beta_i` modulo `n`. Let `\xi` be the sign of the
            permutation that is used for this sorting. Let `\psi` be the sign
            of the permutation that is used to stably sort the list
            `(1, 2, \ldots, ns)` in order of (weakly) increasing remainder of
            `i - 1` modulo `n`. (Notice that `\psi = (-1)^{n(n-1)s(s-1)/4}`.)
            Then, `\mathbf{V}_n(s_\lambda) = \xi \psi \prod_{i = 0}^{n - 1}
            s_{\lambda^{(i)}}`, where
            `(\lambda^{(0)}, \lambda^{(1)}, \ldots, \lambda^{(n - 1)})`
            is the `n`-quotient of `\lambda`.

            INPUT:

            - ``n`` -- positive integer

            OUTPUT:

            The result of applying the `n`-th Verschiebung operator (on the ring of
            symmetric functions) to ``self``.

            EXAMPLES::

                sage: # needs lrcalc_python
                sage: Sym = SymmetricFunctions(ZZ)
                sage: s = Sym.s()
                sage: s[5].verschiebung(2)
                0
                sage: s[6].verschiebung(6)
                s[1]
                sage: s[6,3].verschiebung(3)
                s[2, 1] + s[3]
                sage: s[6,3,1].verschiebung(2)
                -s[3, 2]
                sage: s[3,2,1].verschiebung(1)
                s[3, 2, 1]
                sage: s([]).verschiebung(1)
                s[]
                sage: s([]).verschiebung(4)
                s[]

            TESTS:

            Let us check that this method on the powersum basis gives the
            same result as the implementation in sfa.py on the monomial
            basis::

                sage: # needs lrcalc_python
                sage: Sym = SymmetricFunctions(QQ)
                sage: s = Sym.s(); h = Sym.h()
                sage: all( h(s(lam)).verschiebung(3) == h(s(lam).verschiebung(3))
                ....:      for lam in Partitions(6) )
                True
                sage: all( s(h(lam)).verschiebung(2) == s(h(lam).verschiebung(2))
                ....:      for lam in Partitions(4) )
                True
                sage: all( s(h(lam)).verschiebung(5) == s(h(lam).verschiebung(5))
                ....:      for lam in Partitions(10) )
                True
                sage: all( s(h(lam)).verschiebung(2) == s(h(lam).verschiebung(2))
                ....:      for lam in Partitions(8) )
                True
                sage: all( s(h(lam)).verschiebung(3) == s(h(lam).verschiebung(3))
                ....:      for lam in Partitions(12) )
                True
                sage: all( s(h(lam)).verschiebung(3) == s(h(lam).verschiebung(3))
                ....:      for lam in Partitions(9) )
                True
            """
            # Extra hack for the n == 1 case, since lam.quotient(1)
            # (for lam being a partition) returns a partition rather than
            # a partition tuple.
            if n == 1:
                return self

            parent = self.parent()
            s_coords_of_self = self.monomial_coefficients().items()
            result = parent.zero()
            from sage.combinat.permutation import Permutation
            for (lam, coeff) in s_coords_of_self:
                if len(lam.core(n)) == 0:
                    quotient = lam.quotient(n)
                    quotient_prod = parent.prod(parent(part)
                                                for part in quotient)
                    # Now, compute the sign of quotient_prod in the
                    # n-th Verschiebung of lam.
                    len_lam = len(lam)
                    ns = len_lam + ((- len_lam) % n)
                    s = ns // n   # This is actually ns / n, as we have n | ns.
                    beta_list = lam.beta_numbers(ns)
                    zipped_beta_list = sorted(zip(beta_list, range(1, ns + 1)),
                                              key=lambda a: (-1 - a[0]) % n)
                    # We are using the fact that sort is a stable sort.
                    perm_list = [a[1] for a in zipped_beta_list]
                    if Permutation(perm_list).sign() == 1:
                        minus_sign = False
                    else:
                        minus_sign = True
                    if (n * s * (n-1) * (s-1)) % 8 == 4:
                        minus_sign = not minus_sign
                    if minus_sign:
                        result -= coeff * quotient_prod
                    else:
                        result += coeff * quotient_prod
            return result

        def expand(self, n, alphabet='x'):
            """
            Expand the symmetric function ``self`` as a symmetric polynomial
            in ``n`` variables.

            INPUT:

            - ``n`` -- nonnegative integer

            - ``alphabet`` -- (default: ``'x'``) a variable for the expansion

            OUTPUT:

            A monomial expansion of ``self`` in the `n` variables
            labelled by ``alphabet``.

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ).s()
                sage: a = s([2,1])
                sage: a.expand(2)
                x0^2*x1 + x0*x1^2
                sage: a.expand(3)
                x0^2*x1 + x0*x1^2 + x0^2*x2 + 2*x0*x1*x2 + x1^2*x2 + x0*x2^2 + x1*x2^2
                sage: a.expand(4)
                x0^2*x1 + x0*x1^2 + x0^2*x2 + 2*x0*x1*x2 + x1^2*x2 + x0*x2^2 + x1*x2^2 + x0^2*x3 + 2*x0*x1*x3 + x1^2*x3 + 2*x0*x2*x3 + 2*x1*x2*x3 + x2^2*x3 + x0*x3^2 + x1*x3^2 + x2*x3^2
                sage: a.expand(2, alphabet='y')
                y0^2*y1 + y0*y1^2
                sage: a.expand(2, alphabet=['a','b'])
                a^2*b + a*b^2
                sage: s([1,1,1,1]).expand(3)
                0
                sage: (s([]) + 2*s([1])).expand(3)
                2*x0 + 2*x1 + 2*x2 + 1
                sage: s([1]).expand(0)
                0
                sage: (3*s([])).expand(0)
                3
            """
            condition = lambda part: len(part) > n
            return self._expand(condition, n, alphabet)

        def principal_specialization(self, n=infinity, q=None):
            r"""
            Return the principal specialization of a symmetric function.

            The *principal specialization* of order `n` at `q`
            is the ring homomorphism `ps_{n,q}` from the ring of
            symmetric functions to another commutative ring `R`
            given by `x_i \mapsto q^{i-1}` for `i \in \{1,\dots,n\}`
            and `x_i \mapsto 0` for `i > n`.
            Here, `q` is a given element of `R`, and we assume that
            the variables of our symmetric functions are
            `x_1, x_2, x_3, \ldots`.
            (To be more precise, `ps_{n,q}` is a `K`-algebra
            homomorphism, where `K` is the base ring.)
            See Section 7.8 of [EnumComb2]_.

            The *stable principal specialization* at `q` is the ring
            homomorphism `ps_q` from the ring of symmetric functions
            to another commutative ring `R` given by
            `x_i \mapsto q^{i-1}` for all `i`.
            This is well-defined only if the resulting infinite sums
            converge; thus, in particular, setting `q = 1` in the
            stable principal specialization is an invalid operation.

            INPUT:

            - ``n`` -- (default: ``infinity``) a nonnegative integer or
              ``infinity``, specifying whether to compute the principal
              specialization of order ``n`` or the stable principal
              specialization.

            - ``q`` -- (default: ``None``) the value to use for `q`; the
              default is to create a ring of polynomials in ``q``
              (or a field of rational functions in ``q``) over the
              given coefficient ring.

            For `q=1` we use the formula from Corollary 7.21.4 of [EnumComb2]_:

            .. MATH::

                ps_{n,1}(s_\lambda) = \prod_{u\in\lambda} (n+c(u)) / h(u),

            where `h(u)` is the hook length of a cell `u` in `\lambda`,
            and where `c(u)` is the content of a cell `u` in `\lambda`.

            For `n=infinity` we use the formula from Corollary 7.21.3 of [EnumComb2]_

            .. MATH::

                ps_q(s_\lambda) = q^{\sum_i (i-1)\lambda_i} / \prod_{u\in\lambda} (1-q^{h(u)}).

            Otherwise, we use the formula from Theorem 7.21.2 of [EnumComb2]_,

            .. MATH::

                ps_{n,q}(s_\lambda) = q^{\sum_i (i-1)\lambda_i}
                                      \prod_{u\in\lambda} (1-q^{n+c(u)})/(1-q^{h(u)}).

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ).s()
                sage: x = s[2]
                sage: x.principal_specialization(3)
                q^4 + q^3 + 2*q^2 + q + 1

                sage: x = 3*s[2,2] + 2*s[1] + 1
                sage: x.principal_specialization(3, q=var("q"))                         # needs sage.symbolic
                3*(q^4 - 1)*(q^3 - 1)*q^2/((q^2 - 1)*(q - 1)) + 2*(q^3 - 1)/(q - 1) + 1

                sage: x.principal_specialization(q=var("q"))                            # needs sage.symbolic
                -2/(q - 1) + 3*q^2/((q^3 - 1)*(q^2 - 1)^2*(q - 1)) + 1

            TESTS::

                sage: s.zero().principal_specialization(3)
                0
            """
            def get_variable(ring, name):
                try:
                    ring(name)
                except TypeError:
                    from sage.rings.polynomial.polynomial_ring_constructor import (
                        PolynomialRing,
                    )
                    return PolynomialRing(ring, name).gen()
                else:
                    raise ValueError("the variable %s is in the base ring, pass it explicitly" % name)

            if q is None:
                q = get_variable(self.base_ring(), 'q')
            if q == 1:
                if n == infinity:
                    raise ValueError("the stable principal specialization at q=1 is not defined")
                f = lambda partition: (prod(n+j-i for (i, j) in partition.cells())
                                       // prod(h for h in partition.hooks()))
            elif n == infinity:
                f = lambda partition: (q**sum(i*part for i, part in enumerate(partition))
                                       / prod(1-q**h for h in partition.hooks()))
            else:
                from sage.rings.integer_ring import ZZ
                ZZq = PolynomialRing(ZZ, "q")
                q_lim = ZZq.gen()

                def f(partition):
                    if n < len(partition):
                        return 0
                    power = q**sum(i * part for i, part in enumerate(partition))
                    denom = prod(1 - q**h for h in partition.hooks())
                    try:
                        ~denom
                        rational = (power
                                    * prod(1-q**(n+j-i)
                                           for (i, j) in partition.cells())
                                    / denom)
                        return q.parent()(rational)
                    except (ZeroDivisionError, NotImplementedError, TypeError):
                        # If denom is not invertible, we need to do the
                        # computation with universal coefficients instead:
                        quotient = ZZq((prod(1-q_lim**(n+j-i)
                                             for (i, j) in partition.cells()))
                                    / prod(1-q_lim**h for h in partition.hooks()))
                        return power * quotient.subs({q_lim: q})

            return self.parent()._apply_module_morphism(self, f, q.parent())

        def exponential_specialization(self, t=None, q=1):
            r"""
            Return the exponential specialization of a
            symmetric function (when `q = 1`), or the
            `q`-exponential specialization (when `q \neq 1`).

            The *exponential specialization* `ex` at `t` is a
            `K`-algebra homomorphism from the `K`-algebra of
            symmetric functions to another `K`-algebra `R`.
            It is defined whenever the base ring `K` is a
            `\QQ`-algebra and `t` is an element of `R`.
            The easiest way to define it is by specifying its
            values on the powersum symmetric functions to be
            `p_1 = t` and `p_n = 0` for `n > 1`.
            Equivalently, on the homogeneous functions it is
            given by `ex(h_n) = t^n / n!`; see Proposition 7.8.4 of
            [EnumComb2]_.

            By analogy, the `q`-exponential specialization is a
            `K`-algebra homomorphism from the `K`-algebra of
            symmetric functions to another `K`-algebra `R` that
            depends on two elements `t` and `q` of `R` for which
            the elements `1 - q^i` for all positive integers `i`
            are invertible.
            It can be defined by specifying its values on the
            complete homogeneous symmetric functions to be

            .. MATH::

                ex_q(h_n) = t^n / [n]_q!,

            where `[n]_q!` is the `q`-factorial.  Equivalently, for
            `q \neq 1` and a homogeneous symmetric function `f` of
            degree `n`, we have

            .. MATH::

                ex_q(f) = (1-q)^n t^n ps_q(f),

            where `ps_q(f)` is the stable principal specialization of `f`
            (see :meth:`principal_specialization`).
            (See (7.29) in [EnumComb2]_.)

            The limit of `ex_q` as `q \to 1` is `ex`.

            INPUT:

            - ``t`` -- (default: ``None``) the value to use for `t`;
              the default is to create a ring of polynomials in ``t``

            - ``q`` -- (default: `1`) the value to use for `q`.  If
              ``q`` is ``None``, then a ring (or fraction field) of
              polynomials in ``q`` is created.

            We use the formula in the proof of Corollary 7.21.6 of
            [EnumComb2]_

            .. MATH::

                ex_{q}(s_\lambda) = t^{|\lambda|} q^{\sum_i (i-1)\lambda_i}
                                      / \prod_{u\in\lambda} (1 + q + q^2 + \dots + q^{h(u)-1})

            where `h(u)` is the hook length of a cell `u` in `\lambda`.

            As a limit case, we obtain a formula for `q=1`

            .. MATH::

                ex_{1}(s_\lambda) = f^\lambda t^{|\lambda|} / |\lambda|!

            where `f^\lambda` is the number of standard Young
            tableaux of shape `\lambda`.

            EXAMPLES::

                sage: s = SymmetricFunctions(QQ).s()
                sage: x = s[5,3]
                sage: x.exponential_specialization()
                1/1440*t^8

                sage: x = 5*s[1,1,1] + 3*s[2,1] + 1
                sage: x.exponential_specialization()
                11/6*t^3 + 1

            We also support the `q`-exponential_specialization::

                sage: factor(s[3].exponential_specialization(q=var("q"), t=var("t")))   # needs sage.symbolic
                t^3/((q^2 + q + 1)*(q + 1))

            TESTS::

                sage: s.zero().exponential_specialization()
                0
            """
            def get_variable(ring, name):
                try:
                    ring(name)
                except TypeError:
                    from sage.rings.polynomial.polynomial_ring_constructor import (
                        PolynomialRing,
                    )
                    return PolynomialRing(ring, name).gen()
                else:
                    raise ValueError("the variable %s is in the base ring, pass it explicitly" % name)

            if q == 1:
                if t is None:
                    t = get_variable(self.base_ring(), 't')

                def f(partition):
                    n = partition.size()
                    return (StandardTableaux(partition).cardinality()
                            * t**n / factorial(n))

                return self.parent()._apply_module_morphism(self, f, t.parent())

            if q is None and t is None:
                q = get_variable(self.base_ring(), 'q')
                t = get_variable(q.parent(), 't')
            elif q is None:
                q = get_variable(t.parent(), 'q')
            elif t is None:
                t = get_variable(q.parent(), 't')

            f = lambda partition: (t**partition.size()
                                   * q**sum(i*part for i, part in enumerate(partition))
                                   / prod(sum(q**i for i in range(h)) for h in partition.hooks()))

            return self.parent()._apply_module_morphism(self, f, t.parent())


# Backward compatibility for unpickling
from sage.misc.persist import register_unpickle_override

register_unpickle_override('sage.combinat.sf.schur',
                           'SymmetricFunctionAlgebraElement_schur',
                           SymmetricFunctionAlgebra_schur.Element)



# sage_setup: distribution = sagemath-combinat
r"""
The Hillman-Grassl correspondence

This module implements weak reverse plane partitions and four
correspondences on them: the Hillman-Grassl correspondence and
its inverse, as well as the Sulzgruber correspondence and its
inverse (the Pak correspondence).

Fix a partition `\lambda`
(see :meth:`~sage.combinat.partition.Partition`).
We draw all partitions and tableaux in English notation.

A `\lambda`-*array* will mean a tableau of shape `\lambda` whose
entries are nonnegative integers. (No conditions on the order of
these entries are made. Note that `0` is allowed.)

A *weak reverse plane partition of shape* `\lambda` (short:
`\lambda`-*rpp*) will mean a `\lambda`-array whose entries weakly
increase along each row and weakly increase along each column.
(The name "weak reverse plane partition" comes from Stanley in
[EnumComb2]_ Section 7.22; other authors -- such as Pak
[Sulzgr2017]_, or Hillman and Grassl in [HilGra1976]_ -- just
call it a reverse plane partition.)

The Hillman-Grassl correspondence is a bijection from the
set of `\lambda`-arrays to the set of `\lambda`-rpps.
For its definition, see
:meth:`~sage.combinat.tableau.Tableau.hillman_grassl`;
for its inverse, see
:meth:`~sage.combinat.hillman_grassl.WeakReversePlanePartition.hillman_grassl_inverse`.

The Sulzgruber correspondence `\Phi_\lambda` and the Pak
correspondence `\xi_\lambda` are two further mutually
inverse bijections between the set of
`\lambda`-arrays and the set of `\lambda`-rpps.
They appear (sometimes with different definitions, but
defining the same maps) in [Pak2002]_, [Hopkins2017]_ and
[Sulzgr2017]_. For their definitions, see
:meth:`~sage.combinat.tableau.Tableau.sulzgruber_correspondence` and
:meth:`~sage.combinat.hillman_grassl.WeakReversePlanePartition.pak_correspondence`.

EXAMPLES:

We construct a `\lambda`-rpp for `\lambda = (3, 3, 1)`
(note that `\lambda` needs not be specified explicitly)::

    sage: p = WeakReversePlanePartition([[0, 1, 3], [2, 4, 4], [3]])
    sage: p.parent()
    Weak Reverse Plane Partitions

(This is the example in Section 7.22 of [EnumComb2]_.)

Next, we apply the inverse of the Hillman-Grassl correspondence
to it::

    sage: HGp = p.hillman_grassl_inverse(); HGp
    [[1, 2, 0], [1, 0, 1], [1]]
    sage: HGp.parent()
    Tableaux

This is a `\lambda`-array, encoded as a tableau. We can
recover our original `\lambda`-rpp from it using the
Hillman-Grassl correspondence::

    sage: HGp.hillman_grassl() == p
    True

We can also apply the Pak correspondence to our rpp::

    sage: Pp = p.pak_correspondence(); Pp
    [[2, 0, 1], [0, 2, 0], [1]]
    sage: Pp.parent()
    Tableaux

This is undone by the Sulzgruber correspondence::

    sage: Pp.sulzgruber_correspondence() == p
    True

These four correspondences can also be accessed as standalone
functions (:meth:`hillman_grassl_inverse`, :meth:`hillman_grassl`,
:meth:`pak_correspondence` and :meth:`sulzgruber_correspondence`)
that transform lists of lists into lists of lists;
this may be more efficient. For example, the above computation
of ``HGp`` can also be obtained as follows::

    sage: from sage.combinat.hillman_grassl import hillman_grassl_inverse
    sage: HGp_bare = hillman_grassl_inverse([[0, 1, 3], [2, 4, 4], [3]])
    sage: HGp_bare
    [[1, 2, 0], [1, 0, 1], [1]]
    sage: isinstance(HGp_bare, list)
    True

REFERENCES:

- [Gans1981]_
- [HilGra1976]_
- [EnumComb2]_
- [Pak2002]_
- [Sulzgr2017]_
- [Hopkins2017]_

AUTHORS:

- Darij Grinberg and Tom Roby (2018): Initial implementation
"""

# ****************************************************************************
#       Copyright (C) 2018 Darij Grinberg <darijgrinberg@gmail.com>,
#                     2018 Tom Roby <tomrobyuconn@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.tableau import Tableau, Tableaux
from sage.categories.sets_cat import Sets
from sage.combinat.combinatorial_map import combinatorial_map


class WeakReversePlanePartition(Tableau):
    r"""
    A weak reverse plane partition (short: rpp).

    A weak reverse plane partition is a tableau with nonnegative
    entries that are weakly increasing in each row and weakly
    increasing in each column.

    EXAMPLES::

        sage: x = WeakReversePlanePartition([[0, 1, 1], [0, 1, 3], [1, 2, 2], [1, 2, 3], [2]]); x
        [[0, 1, 1], [0, 1, 3], [1, 2, 2], [1, 2, 3], [2]]
        sage: x.pp()
          0  1  1
          0  1  3
          1  2  2
          1  2  3
          2
        sage: x.shape()
        [3, 3, 3, 3, 1]
    """
    @staticmethod
    def __classcall_private__(cls, r):
        r"""
        Return an rpp object.

        EXAMPLES::

            sage: WeakReversePlanePartition([[1, 2], [1, 3], [1]])
            [[1, 2], [1, 3], [1]]

        TESTS::

            sage: a1 = [[1, 2], [1, 3], [1]]
            sage: a2 = [(1, 2), (1, 3), (1,)]
            sage: A1 = WeakReversePlanePartition(a1)
            sage: A2 = WeakReversePlanePartition(a2)
            sage: A3 = WeakReversePlanePartition(A1)
            sage: A4 = Tableau(A1)
            sage: A1 == A2 == A3 == A4
            True
        """
        try:
            r = list(map(tuple, r))
        except TypeError:
            raise TypeError("r must be a list of positive integers")
        return WeakReversePlanePartitions()(r)

    def __init__(self, parent, t):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: R = WeakReversePlanePartition([[0, 1, 2], [0, 2]])
            sage: TestSuite(R).run()
        """
        if not isinstance(t, Tableau):
            t = [list(row) for row in t]
        else:
            t = list(t)

        Tableau.__init__(self, parent, t)

    @combinatorial_map(order=2, name='conjugate')
    def conjugate(self):
        """
        Return the conjugate of ``self``.

        EXAMPLES::

            sage: c = WeakReversePlanePartition([[1,1],[1,3],[2]]).conjugate(); c
            [[1, 1, 2], [1, 3]]
            sage: c.parent()
            Weak Reverse Plane Partitions
        """
        C = super().conjugate()
        return WeakReversePlanePartition(C)

    def hillman_grassl_inverse(self):
        r"""
        Return the image of the `\lambda`-rpp ``self`` under the
        inverse of the Hillman-Grassl correspondence (as a
        :class:`~sage.combinat.tableau.Tableau`).

        Fix a partition `\lambda`
        (see :meth:`~sage.combinat.partition.Partition`).
        We draw all partitions and tableaux in English notation.

        A `\lambda`-*array* will mean a tableau of shape `\lambda` whose
        entries are nonnegative integers. (No conditions on the order of
        these entries are made. Note that `0` is allowed.)

        A *weak reverse plane partition of shape* `\lambda` (short:
        `\lambda`-*rpp*) will mean a `\lambda`-array whose entries weakly
        increase along each row and weakly increase along each column.

        The inverse `H^{-1}` of the Hillman-Grassl correspondence (see
        (:meth:`~sage.combinat.tableau.Tableau.hillman_grassl` for the
        latter) sends a `\lambda`-rpp `\pi` to a `\lambda`-array
        `H^{-1}(\pi)` constructed recursively as follows:

        * If all entries of `\pi` are `0`, then `H^{-1}(\pi) = \pi`.

        * Otherwise, let `s` be the index of the leftmost column of `\pi`
          containing a nonzero entry. Write the `\lambda`-array `M`
          as `(m_{i, j})`.

        * Define a sequence `((i_1, j_1), (i_2, j_2), \ldots,
          (i_n, j_n))` of boxes in the diagram of `\lambda` (actually a
          lattice path made of northward and eastward steps) as follows:
          Let `(i_1, j_1)` be the bottommost box in the `s`-th column
          of `\pi`.
          If `(i_k, j_k)` is defined for some `k \geq 1`, then
          `(i_{k+1}, j_{k+1})` is constructed as follows:
          If `q_{i_k - 1, j_k}` is well-defined and equals `q_{i_k, j_k}`,
          then we set `(i_{k+1}, j_{k+1}) = (i_k - 1, j_k)`. Otherwise,
          we set `(i_{k+1}, j_{k+1}) = (i_k, j_k + 1)` if this is still
          a box of `\lambda`. Otherwise, the sequence ends here.

        * Let `\pi'` be the `\lambda`-rpp obtained from `\pi` by
          subtracting `1` from the `(i_k, j_k)`-th entry of `\pi` for each
          `k \in \{1, 2, \ldots, n\}`.

        * Let `N'` be the image `H^{-1}(\pi')` (which is already
          constructed by recursion).
          Then, `H^{-1}(\pi)` is obtained from `N'` by adding `1` to the
          `(i_n, s)`-th entry of `N'`.

        This construction appears in [HilGra1976]_ Section 6 (where
        `\lambda`-arrays are re-encoded as sequences of "hook number
        multiplicities") and [EnumComb2]_ Section 7.22.

        .. SEEALSO::

            :meth:`~sage.combinat.hillman_grassl.hillman_grassl_inverse`
            for the inverse of the Hillman-Grassl correspondence as a
            standalone function.

            :meth:`~sage.combinat.tableau.Tableau.hillman_grassl`
            for the inverse map.

        EXAMPLES::

            sage: a = WeakReversePlanePartition([[2, 2, 4], [2, 3, 4], [3, 5]])
            sage: a.hillman_grassl_inverse()
            [[2, 1, 1], [0, 2, 0], [1, 1]]
            sage: b = WeakReversePlanePartition([[1, 1, 2, 2], [1, 1, 2, 2], [2, 2, 3, 3], [2, 2, 3, 3]])
            sage: B = b.hillman_grassl_inverse(); B
            [[1, 0, 1, 0], [0, 1, 0, 1], [1, 0, 1, 0], [0, 1, 0, 1]]
            sage: b.parent(), B.parent()
            (Weak Reverse Plane Partitions, Tableaux)

        Applying the inverse of the Hillman-Grassl correspondence
        to the transpose of a `\lambda`-rpp `M` yields the same
        result as applying it to `M` and then transposing the
        result ([Gans1981]_ Corollary 3.4)::

            sage: a = WeakReversePlanePartition([[1,3,5],[2,4]])
            sage: aic = a.hillman_grassl_inverse().conjugate()
            sage: aic == a.conjugate().hillman_grassl_inverse()
            True
        """
        return Tableau(hillman_grassl_inverse(list(self)))

    def pak_correspondence(self):
        r"""
        Return the image of the `\lambda`-rpp ``self`` under the Pak
        correspondence (as a :class:`~sage.combinat.tableau.Tableau`).

        See :mod:`~sage.combinat.hillman_grassl`.

        The Pak correspondence is the map `\xi_\lambda`
        from [Sulzgr2017]_ Section 7, and is the map
        `\xi_\lambda` from [Pak2002]_ Section 4.
        It is the inverse of the Sulzgruber correspondence
        (:meth:`sulzgruber_correspondence`).
        The following description of the Pak correspondence follows
        [Hopkins2017]_ (which denotes it by `\mathcal{RSK}^{-1}`):

        Fix a partition `\lambda`
        (see :meth:`~sage.combinat.partition.Partition`).
        We draw all partitions and tableaux in English notation.

        A `\lambda`-*array* will mean a tableau of shape `\lambda` whose
        entries are nonnegative integers. (No conditions on the order of
        these entries are made. Note that `0` is allowed.)

        A *weak reverse plane partition of shape* `\lambda` (short:
        `\lambda`-*rpp*) will mean a `\lambda`-array whose entries weakly
        increase along each row and weakly increase along each column.

        We shall also use the following notation:
        If `(u, v)` is a cell of `\lambda`, and if `\pi` is a
        `\lambda`-rpp, then:

        * the *lower bound* of `\pi` at `(u, v)` (denoted by
          `\pi_{<(u, v)}`) is defined to be
          `\max \{ \pi_{u-1, v} , \pi_{u, v-1} \}`
          (where `\pi_{0, v}` and `\pi_{u, 0}` are understood to
          mean `0`).

        * the *upper bound* of `\pi` at `(u, v)` (denoted by
          `\pi_{>(u, v)}`) is defined to be
          `\min \{ \pi_{u+1, v} , \pi_{u, v+1} \}`
          (where `\pi_{i, j}` is understood to mean `+ \infty`
          if `(i, j)` is not in `\lambda`; thus, the upper
          bound at a corner cell is `+ \infty`).

        * *toggling* `\pi` at `(u, v)` means replacing the entry
          `\pi_{u, v}` of `\pi` at `(u, v)` by
          `\pi_{<(u, v)} + \pi_{>(u, v)} - \pi_{u, v}`
          (this is well-defined as long as `(u, v)` is not a
          corner of `\lambda`).

        Note that every `\lambda`-rpp `\pi` and every cell
        `(u, v)` of `\lambda` satisfy
        `\pi_{<(u, v)} \leq \pi_{u, v} \leq \pi_{>(u, v)}`.
        Note that toggling a `\lambda`-rpp (at a cell that is not
        a corner) always results in a `\lambda`-rpp. Also,
        toggling is an involution).

        Note also that the lower bound of `\pi` at `(u, v)` is
        defined (and finite) even when `(u, v)` is not a cell of
        `\lambda`, as long as both `(u-1, v)` and `(u, v-1)` are
        cells of `\lambda`.

        The Pak correspondence `\Phi_\lambda` sends a `\lambda`-array
        `M = (m_{i, j})` to a `\lambda`-rpp `\Phi_\lambda(M)`. It
        is defined by recursion on `\lambda` (that is, we assume that
        `\Phi_\mu` is already defined for every partition `\mu`
        smaller than `\lambda`), and its definition proceeds as
        follows:

        * If `\lambda = \varnothing`, then `\Phi_\lambda` is the
          obvious bijection sending the only `\varnothing`-array
          to the only `\varnothing`-rpp.

        * Pick any corner `c = (i, j)` of `\lambda`, and let `\mu`
          be the result of removing this corner `c` from the partition
          `\lambda`. (The exact choice of `c` is immaterial.)

        * Let `M'` be what remains of `M` when the corner cell `c`
          is removed.

        * Let `\pi' = \Phi_\mu(M')`.

        * For each positive integer `k` such that `(i-k, j-k)` is a
          cell of `\lambda`, toggle `\pi'` at `(i-k, j-k)`.
          (All these togglings commute, so the order in which they
          are made is immaterial.)

        * Extend the `\mu`-rpp `\pi'` to a `\lambda`-rpp `\pi` by
          adding the cell `c` and writing the number
          `m_{i, j} - \pi'_{<(i, j)}` into this cell.

        * Set `\Phi_\lambda(M) = \pi`.

        .. SEEALSO::

            :meth:`~sage.combinat.hillman_grassl.pak_correspondence`
            for the Pak correspondence as a standalone function.

            :meth:`~sage.combinat.tableau.Tableau.sulzgruber_correspondence`
            for the inverse map.

        EXAMPLES::

            sage: a = WeakReversePlanePartition([[1, 2, 3], [1, 2, 3], [2, 4, 4]])
            sage: A = a.pak_correspondence(); A
            [[1, 0, 2], [0, 2, 0], [1, 1, 0]]
            sage: a.parent(), A.parent()
            (Weak Reverse Plane Partitions, Tableaux)

        Applying the Pak correspondence to the transpose of a
        `\lambda`-rpp `M` yields the same result as applying it to
        `M` and then transposing the result::

            sage: a = WeakReversePlanePartition([[1,3,5],[2,4]])
            sage: acc = a.pak_correspondence().conjugate()
            sage: acc == a.conjugate().pak_correspondence()
            True
        """
        return Tableau(pak_correspondence(list(self)))


class WeakReversePlanePartitions(Tableaux):
    r"""
    The set of all weak reverse plane partitions.
    """
    @staticmethod
    def __classcall_private__(cls, shape=None, **kwds):
        """
        Normalize input to ensure a unique representation and
        return the correct class based on input.

        The ``shape`` parameter is currently not implemented.

        EXAMPLES::

            sage: S1 = WeakReversePlanePartitions()
            sage: S2 = WeakReversePlanePartitions()
            sage: S1 is S2
            True

            sage: S1 = WeakReversePlanePartitions([4, 2, 2, 1])  # not tested (not implemented)
            sage: S2 = WeakReversePlanePartitions((4, 2, 2, 1))  # not tested (not implemented)
            sage: S1 is S2  # not tested (not implemented)
            True
        """
        if shape is not None:
            raise NotImplementedError("shape cannot be specified")
            # from sage.combinat.partition import Partition
            # return RibbonShapedTableaux_shape(Partition(shape))

        return super().__classcall__(cls, **kwds)

    def __init__(self):
        """
        Initialize ``self``.

        EXAMPLES::

            sage: S = WeakReversePlanePartitions()
            sage: TestSuite(S).run()
        """
        Tableaux.__init__(self, category=Sets())

    def _repr_(self):
        """
        TESTS::

            sage: WeakReversePlanePartitions()
            Weak Reverse Plane Partitions
        """
        return "Weak Reverse Plane Partitions"

    Element = WeakReversePlanePartition

    def _an_element_(self):
        r"""
        Return a particular element of the class.

        TESTS::

            sage: T = WeakReversePlanePartitions()
            sage: T.an_element()
            [[0, 0, 1, 2], [0, 1, 1], [0], [2]]
        """
        return self.element_class(self, [[0, 0, 1, 2], [0, 1, 1], [0], [2]])


def transpose(M):
    r"""
    Return the transpose of a `\lambda`-array.

    The transpose of a `\lambda`-array `(m_{i, j})` is the
    `\lambda^t`-array `(m_{j, i})`
    (where `\lambda^t` is the conjugate of the partition
    `\lambda`).

    EXAMPLES::

        sage: from sage.combinat.hillman_grassl import transpose
        sage: transpose([[1, 2, 3], [4, 5]])
        [[1, 4], [2, 5], [3]]
        sage: transpose([[5, 0, 3], [4, 1, 0], [7]])
        [[5, 4, 7], [0, 1], [3, 0]]

    TESTS::

        sage: transpose(((2, 1), (3,)))
        [[2, 3], [1]]
        sage: transpose([])
        []
        sage: transpose(WeakReversePlanePartition([[1, 2, 3], [4, 5]]))
        [[1, 4], [2, 5], [3]]
        sage: transpose(WeakReversePlanePartition([]))
        []
    """
    if not M:  # empty array
        return []
    l = len(M[0])
    res = []
    for j in range(l):
        lj = []
        for r in M:
            if len(r) <= j:
                break
            lj.append(r[j])
        res.append(lj)
    return res


def hillman_grassl(M):
    r"""
    Return the image of the `\lambda`-array ``M``
    under the Hillman-Grassl correspondence.

    The Hillman-Grassl correspondence is a bijection
    between the tableaux with nonnegative entries
    (otherwise arbitrary) and the weak reverse plane
    partitions with nonnegative entries.
    This bijection preserves the shape of the
    tableau. See :mod:`~sage.combinat.hillman_grassl`.

    See :meth:`~sage.combinat.tableau.Tableau.hillman_grassl`
    for a description of this map.

    .. SEEALSO::

        :meth:`hillman_grassl_inverse`

    EXAMPLES::

        sage: from sage.combinat.hillman_grassl import hillman_grassl
        sage: hillman_grassl([[2, 1, 1], [0, 2, 0], [1, 1]])
        [[2, 2, 4], [2, 3, 4], [3, 5]]
        sage: hillman_grassl([[1, 2, 0], [1, 0, 1], [1]])
        [[0, 1, 3], [2, 4, 4], [3]]
        sage: hillman_grassl([])
        []
        sage: hillman_grassl([[3, 1, 2]])
        [[3, 4, 6]]
        sage: hillman_grassl([[2, 2, 0], [1, 1, 1], [1]])
        [[1, 2, 4], [3, 5, 5], [4]]
        sage: hillman_grassl([[1, 1, 1, 1]]*3)
        [[1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6]]

    TESTS::

        sage: hillman_grassl(((2, 2, 0), (1, 1, 1), (1,)))
        [[1, 2, 4], [3, 5, 5], [4]]
    """
    lam = [len(row) for row in M]
    l = len(lam)
    Mt = transpose(M)
    # We won't touch M from now on; we'll only modify Mt
    # in place.
    hook_mults = []
    for j, col_j in enumerate(Mt):
        col_j_hook_mults = []
        for r, entry in enumerate(col_j):
            if entry != 0:
                col_j_hook_mults += [(r, j)] * entry
        hook_mults += reversed(col_j_hook_mults)
    res = [[0] * rowlen for rowlen in lam]
    for (r, s) in reversed(hook_mults):
        i = r
        j = lam[r] - 1
        while True:
            old = res[i][j]
            res[i][j] += 1
            if i + 1 != l and j < lam[i + 1] and old == res[i + 1][j]:
                i += 1
            else:
                if j == s:
                    break
                j -= 1
    return res


def hillman_grassl_inverse(M):
    r"""
    Return the image of the `\lambda`-rpp ``M`` under the
    inverse of the Hillman-Grassl correspondence.

    See :mod:`~sage.combinat.hillman_grassl`.

    See
    :meth:`~sage.combinat.hillman_grassl.WeakReversePlanePartition.hillman_grassl_inverse`
    for a description of this map.

    .. SEEALSO::

        :meth:`hillman_grassl`

    EXAMPLES::

        sage: from sage.combinat.hillman_grassl import hillman_grassl_inverse
        sage: hillman_grassl_inverse([[2, 2, 4], [2, 3, 4], [3, 5]])
        [[2, 1, 1], [0, 2, 0], [1, 1]]
        sage: hillman_grassl_inverse([[0, 1, 3], [2, 4, 4], [3]])
        [[1, 2, 0], [1, 0, 1], [1]]

    Applying the inverse of the Hillman-Grassl correspondence
    to the transpose of a `\lambda`-rpp `M` yields the same
    result as applying it to `M` and then transposing the
    result ([Gans1981]_ Corollary 3.4)::

        sage: hillman_grassl_inverse([[1,3,5],[2,4]])
        [[1, 2, 2], [1, 1]]
        sage: hillman_grassl_inverse([[1,2],[3,4],[5]])
        [[1, 1], [2, 1], [2]]
        sage: hillman_grassl_inverse([[1, 2, 3], [1, 2, 3], [2, 4, 4]])
        [[1, 2, 0], [0, 1, 1], [1, 0, 1]]
        sage: hillman_grassl_inverse([[1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6]])
        [[1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]]

    TESTS::

        sage: a = [[3], [4], [6]]
        sage: hillman_grassl_inverse(a)
        [[3], [1], [2]]
        sage: a
        [[3], [4], [6]]

        sage: hillman_grassl_inverse(((1,2),(3,4),(5,)))
        [[1, 1], [2, 1], [2]]
    """
    lam = [len(row) for row in M]
    res = [[0] * rowlen for rowlen in lam]
    Mt = transpose(M)
    # We won't touch M from now on; we'll only modify Mt
    # in place.
    while True:
        for j, col_j in enumerate(Mt):
            if all(entry == 0 for entry in col_j):
                continue
            else:
                break
        else:  # all entries of Mt are 0.
            break
        # Now, j is the index of the leftmost nonzero column of
        # the array.
        s = j
        i = len(col_j) - 1  # We already have j = s.
        while True:
            old = col_j[i]
            col_j[i] -= 1
            if i > 0 and old == col_j[i - 1]:
                i -= 1
            else:
                j += 1
                if j == lam[i]:
                    break
                col_j = Mt[j]
        res[i][s] += 1
    return res


def sulzgruber_correspondence(M):
    r"""
    Return the image of a `\lambda`-array ``M``
    under the Sulzgruber correspondence.

    The Sulzgruber correspondence is the map `\Phi_\lambda`
    from [Sulzgr2017]_ Section 7, and is the map
    `\xi_\lambda^{-1}` from [Pak2002]_ Section 5.
    It is denoted by `\mathcal{RSK}` in [Hopkins2017]_.
    It is the inverse of the Pak correspondence
    (:meth:`pak_correspondence`).

    See :meth:`~sage.combinat.tableau.Tableau.sulzgruber_correspondence`
    for a description of this map.

    EXAMPLES::

        sage: from sage.combinat.hillman_grassl import sulzgruber_correspondence
        sage: sulzgruber_correspondence([[1, 0, 2], [0, 2, 0], [1, 1, 0]])
        [[1, 2, 3], [1, 2, 3], [2, 4, 4]]
        sage: sulzgruber_correspondence([[1, 1, 2], [0, 1, 0], [3, 0, 0]])
        [[1, 1, 4], [2, 3, 4], [4, 4, 4]]
        sage: sulzgruber_correspondence([[1, 0, 2], [0, 2, 0], [1, 1]])
        [[0, 2, 3], [1, 3, 3], [2, 4]]
        sage: sulzgruber_correspondence([[0, 2, 2], [1, 1], [2]])
        [[1, 2, 4], [1, 3], [3]]
        sage: sulzgruber_correspondence([[1, 1, 1, 1]]*3)
        [[1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6]]

    The Sulzgruber correspondence can actually be
    extended (by the same definition) to arrays
    of nonnegative reals rather than nonnegative
    integers. This implementation supports this::

        sage: sulzgruber_correspondence([[1/2, 0, 1], [0, 1, 0], [1/2, 1/2]])
        [[0, 1, 3/2], [1/2, 3/2, 3/2], [1, 2]]

    TESTS::

        sage: sulzgruber_correspondence([])
        []
        sage: sulzgruber_correspondence(((0, 2, 2), (1, 1), (2,)))
        [[1, 2, 4], [1, 3], [3]]
    """
    lam = [len(row) for row in M]
    l = len(lam)
    if l == 0:
        return []
    # Finding a corner of lam...
    lam_0 = lam[0]
    for i, lam_i in enumerate(lam):
        if lam_i != lam_0:
            i -= 1
            break
    j = lam_0 - 1
    # Now, i is the index of the last row of ``M`` that
    # has the same length as the first row; hence, (i, j)
    # is a corner of lam.
    x = M[i][j]
    N = [list(row) for row in M]

    # remove the corner (i, j):
    N[i].pop()
    if not N[i]:
        N.pop()

    N = sulzgruber_correspondence(N)

    # toggling and inserting the new entry:
    for k in range(min(i, j) + 1):
        u = i - k
        v = j - k
        if u > 0 and v > 0:
            lower_bound = max(N[u - 1][v], N[u][v - 1])
        elif u > 0:
            lower_bound = N[u - 1][v]
        elif v > 0:
            lower_bound = N[u][v - 1]
        else:
            lower_bound = 0
        if k > 0:
            val = N[u][v]
            upper_bound = min(N[u + 1][v], N[u][v + 1])
            N[u][v] = lower_bound + upper_bound - val
        else:
            if len(N) <= u:
                N.append([])
            N[u].append(lower_bound + x)

    return N


def pak_correspondence(M, copy=True):
    r"""
    Return the image of a `\lambda`-rpp ``M``
    under the Pak correspondence.

    The Pak correspondence is the map `\xi_\lambda`
    from [Sulzgr2017]_ Section 7, and is the map
    `\xi_\lambda` from [Pak2002]_ Section 4.
    It is the inverse of the Sulzgruber correspondence
    (:meth:`sulzgruber_correspondence`).

    See
    :meth:`~sage.combinat.hillman_grassl.WeakReversePlanePartition.pak_correspondence`
    for a description of this map.

    INPUT:

    - ``copy`` -- boolean (default: ``True``); if set to ``False``, the
      algorithm will mutate the input (but be more efficient)

    EXAMPLES::

        sage: from sage.combinat.hillman_grassl import pak_correspondence
        sage: pak_correspondence([[1, 2, 3], [1, 2, 3], [2, 4, 4]])
        [[1, 0, 2], [0, 2, 0], [1, 1, 0]]
        sage: pak_correspondence([[1, 1, 4], [2, 3, 4], [4, 4, 4]])
        [[1, 1, 2], [0, 1, 0], [3, 0, 0]]
        sage: pak_correspondence([[0, 2, 3], [1, 3, 3], [2, 4]])
        [[1, 0, 2], [0, 2, 0], [1, 1]]
        sage: pak_correspondence([[1, 2, 4], [1, 3], [3]])
        [[0, 2, 2], [1, 1], [2]]
        sage: pak_correspondence([[1, 2, 3, 4], [2, 3, 4, 5], [3, 4, 5, 6]])
        [[1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]]

    The Pak correspondence can actually be
    extended (by the same definition) to "rpps"
    of nonnegative reals rather than nonnegative
    integers. This implementation supports this::

        sage: pak_correspondence([[0, 1, 3/2], [1/2, 3/2, 3/2], [1, 2]])
        [[1/2, 0, 1], [0, 1, 0], [1/2, 1/2]]

    TESTS::

        sage: pak_correspondence([])
        []
        sage: pak_correspondence(((1, 2, 4), (1, 3), (3,)))
        [[0, 2, 2], [1, 1], [2]]

        sage: a = [[0, 2, 3], [1, 3, 3], [2, 4]]
        sage: pak_correspondence(a)
        [[1, 0, 2], [0, 2, 0], [1, 1]]
        sage: a
        [[0, 2, 3], [1, 3, 3], [2, 4]]
        sage: pak_correspondence(a, copy=False)
        [[1, 0, 2], [0, 2, 0], [1, 1]]
        sage: a
        []
    """
    if not M:
        return []
    lam = [len(row) for row in M]
    # Finding a corner of lam...
    lam_0 = lam[0]
    for i, lam_i in enumerate(lam):
        if lam_i != lam_0:
            i -= 1
            break
    j = lam_0 - 1
    # Now, i is the index of the last row of ``M`` that
    # has the same length as the first row; hence, (i, j)
    # is a corner of lam.
    x = M[i][j]

    if copy:
        N = [list(row) for row in M]
        # make a deep copy of M to avoid vandalizing M
    else:
        N = M

    # toggling and obtaining x (the corner entry of the output):
    for k in range(min(i, j) + 1):
        u = i - k
        v = j - k
        if u > 0 and v > 0:
            lower_bound = max(N[u - 1][v], N[u][v - 1])
        elif u > 0:
            lower_bound = N[u - 1][v]
        elif v > 0:
            lower_bound = N[u][v - 1]
        else:
            lower_bound = 0
        if k > 0:
            val = N[u][v]
            upper_bound = min(N[u + 1][v], N[u][v + 1])
            N[u][v] = lower_bound + upper_bound - val
        else:
            x -= lower_bound

    # remove the corner (i, j):
    N[i].pop()
    if not N[i]:
        N.pop()

    N = pak_correspondence(N, copy=False)
    if len(N) <= i:
        N.append([])
    N[i].append(x)
    return N



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.libs.pari sage.modules
r"""
Splitting Algebras

*Splitting algebras* have been considered by Dan Laksov, Anders Thorup,
Torsten Ekedahl and others (see references below) in order to study
intersection theory of Grassmann and other flag schemes. Similarly as
*splitting fields* they can be considered as extensions of rings containing
all the roots of a given monic polynomial over that ring under the
assumption that its Galois group is the symmetric group of order equal
to the polynomial's degree.

Thus they can be used as a tool to express elements of a ring generated by
`n` indeterminates in terms of symmetric functions in these indeterminates.

This realization of splitting algebras follows the approach of a recursive
quotient ring construction splitting off some linear factor of the
polynomial in each recursive step. Accordingly it is inherited from
:class:`PolynomialQuotientRing_domain`.

AUTHORS:

- Sebastian Oehms (April 2020): initial version
"""

# ****************************************************************************
#       Copyright (C) 2020 Sebastian Oehms <seb.oehms@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from warnings import warn

from sage.misc.verbose import verbose
from sage.misc.cachefunc import cached_method
from sage.rings.polynomial.polynomial_quotient_ring import PolynomialQuotientRing_domain
from sage.rings.polynomial.polynomial_quotient_ring_element import PolynomialQuotientRingElement


# ------------------------------------------------------------------------------------------------------------------
# Element class for the splitting algebra
# --------------------------------------------------------------------------------------------------------
class SplittingAlgebraElement(PolynomialQuotientRingElement):
    r"""
    Element class for :class:`SplittingAlgebra`.

    EXAMPLES::

        sage: from sage.algebras.splitting_algebra import SplittingAlgebra
        sage: cp6 = cyclotomic_polynomial(6)
        sage: CR6.<e6> = SplittingAlgebra(cp6)
        sage: type(e6)
        <class 'sage.algebras.splitting_algebra.SplittingAlgebra_with_category.element_class'>

        sage: type(CR6(5))
        <class 'sage.algebras.splitting_algebra.SplittingAlgebra_with_category.element_class'>
    """
    def __invert__(self):
        r"""
        Return the inverse of ``self``.

        Support inversion of special elements attached to the construction
        of the parent and which are recorded in the list
        ``self.parent()._invertible_elements``.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: CR3.<e3> = SplittingAlgebra(cyclotomic_polynomial(3))
            sage: ~e3
            -e3 - 1
            sage: ~(e3 + 5)
            Traceback (most recent call last):
            ...
            ArithmeticError: element is non-invertible
        """
        inv_elements = self.parent()._invertible_elements
        if self in inv_elements:
            return inv_elements[self]

        return super().__invert__()

    def is_unit(self):
        r"""
        Return ``True`` if ``self`` is invertible.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: CR3.<e3> = SplittingAlgebra(cyclotomic_polynomial(3))
            sage: e3.is_unit()
            True
        """
        inv_elements = self.parent()._invertible_elements
        if self in inv_elements:
            return True

        return super().is_unit()

    def monomial_coefficients(self, copy=True):
        r"""
        Return the dictionary of ``self`` according to its lift to the cover.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: CR3.<e3> = SplittingAlgebra(cyclotomic_polynomial(3))
            sage: f = e3 + 42
            sage: f.monomial_coefficients()
            {0: 42, 1: 1}

        ``dict`` is an alias::

            sage: f.dict()
            {0: 42, 1: 1}
        """
        return self.lift().monomial_coefficients(copy=copy)

    dict = monomial_coefficients

# ------------------------------------------------------------------------------------------------------------------
# Parent class of the splitting algebra
# --------------------------------------------------------------------------------------------------------


class SplittingAlgebra(PolynomialQuotientRing_domain):
    r"""
    For a given monic polynomial `p(t)` of degree `n` over a commutative
    ring `R`, the splitting algebra is the universal `R`-algebra in which
    `p(t)` has `n` roots, or, more precisely, over which `p(t)` factors,

    .. MATH::

        p(t) = (t - \xi_1) \cdots (t - \xi_n).

    This class creates an algebra as extension over the base ring of a
    given polynomial `p` such that `p` splits into linear factors over
    that extension. It is assumed (and not checked in general) that the
    Galois group of `p` is the symmetric Group `S(n)`. The construction
    is recursive (following [LT2012]_, 1.3).

    INPUT:

    - ``monic_polynomial`` -- the monic polynomial which should be split
    - ``names`` -- names for the indeterminates to be adjoined to the
      base ring of ``monic_polynomial``
    - ``warning`` -- boolean (default: ``True``); can be used (by setting to
      ``False``) to suppress a warning which will be thrown whenever it cannot
      be checked that the Galois group of ``monic_polynomial`` is maximal

    EXAMPLES::

        sage: from sage.algebras.splitting_algebra import SplittingAlgebra
        sage: Lc.<w> = LaurentPolynomialRing(ZZ)
        sage: PabLc.<u,v> = Lc[]; t = polygen(PabLc)
        sage: S.<x, y> = SplittingAlgebra(t^3 - u*t^2 + v*t - w)
        doctest:...: UserWarning: Assuming x^3 - u*x^2 + v*x - w to have maximal
                                  Galois group!

        sage: roots = S.splitting_roots(); roots
        [x, y, -y - x + u]
        sage: all(t^3 -u*t^2 +v*t -w == 0 for t in roots)
        True
        sage: xi = ~x; xi
        (w^-1)*x^2 + ((-w^-1)*u)*x + (w^-1)*v
        sage: ~xi == x
        True
        sage: ~y
        ((-w^-1)*x)*y + (-w^-1)*x^2 + ((w^-1)*u)*x
        sage: zi = ((w^-1)*x)*y; ~zi
        -y - x + u

        sage: cp3 = cyclotomic_polynomial(3).change_ring(GF(5))
        sage: CR3.<e3> = SplittingAlgebra(cp3)
        sage: CR3.is_field()
        True
        sage: CR3.cardinality()
        25
        sage: F.<a> = cp3.splitting_field()
        sage: F.cardinality()
        25
        sage: E3 = cp3.change_ring(F).roots()[0][0]; E3
        3*a + 3
        sage: f = CR3.hom([E3]); f
        Ring morphism:
          From: Splitting Algebra of x^2 + x + 1
                with roots [e3, 4*e3 + 4]
                over Finite Field of size 5
          To:   Finite Field in a of size 5^2
          Defn: e3 |--> 3*a + 3

    REFERENCES:

    - [EL2002]_
    - [Lak2010]_
    - [Tho2011]_
    - [LT2012]_
    """
    Element = SplittingAlgebraElement

    def __init__(self, monic_polynomial, names='X', iterate=True, warning=True):
        r"""
        Python constructor.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: Lw.<w> = LaurentPolynomialRing(ZZ)
            sage: PuvLw.<u,v> = Lw[]; t = polygen(PuvLw)
            sage: S.<x, y> = SplittingAlgebra(t^3 - u*t^2 + v*t - w, warning=False)
            sage: TestSuite(S).run()
        """

        # ---------------------------------------------------------------------------------
        # checking input parameters
        # ---------------------------------------------------------------------------------

        base_ring = monic_polynomial.base_ring()
        if not monic_polynomial.is_monic():
            raise ValueError("given polynomial must be monic")
        deg = monic_polynomial.degree()

        from sage.structure.category_object import normalize_names
        self._root_names = normalize_names(deg-1, names)
        root_names = list(self._root_names)
        verbose("Create splitting algebra to base ring %s and polynomial %s (%s %s)"
                % (base_ring, monic_polynomial, iterate, warning))

        self._defining_polynomial = monic_polynomial
        self._iterate = iterate

        try:
            if not base_ring.is_integral_domain():
                raise TypeError("base_ring must be an integral domain")
        except NotImplementedError:
            from sage.categories.rings import Rings
            if base_ring not in Rings():
                raise TypeError("base_ring must be a ring")
            if warning:
                warn('Assuming %s to be an integral domain!' % base_ring)

        if deg < 1:
            raise ValueError("the degree of the polynomial must positive")

        self._splitting_roots = []
        self._coefficients_list = []
        self._invertible_elements = {}

        if isinstance(base_ring, SplittingAlgebra):
            self._invertible_elements = base_ring._invertible_elements

        # ------------------------------------------------------------------------------------
        # taking next root_name
        # ------------------------------------------------------------------------------------
        root_name = root_names[0]
        p = monic_polynomial.change_variable_name(root_name)
        P = p.parent()

        self._set_modulus_irreducible_ = False
        try:
            if not p.is_irreducible():
                raise ValueError("monic_polynomial must be irreducible")
        except (NotImplementedError, AttributeError):
            # assuming this has been checked mathematically before
            self._set_modulus_irreducible_ = True
            if warning:
                warn('Assuming %s to have maximal Galois group!' % (monic_polynomial))
                warning = False # one warning must be enough

        verbose("P %s defined:" % (P))

        if deg > 2 and iterate:
            # ------------------------------------------------------------------------------------
            # successive solution via recursion (on base_ring_step)
            # ------------------------------------------------------------------------------------
            base_ring_step = SplittingAlgebra(monic_polynomial, tuple(root_names), iterate=False, warning=False)
            first_root = base_ring_step.gen()

            verbose("base_ring_step %s defined:" % (base_ring_step))

            # -------------------------------------------------------------
            # splitting first root off
            # -------------------------------------------------------------
            from copy import copy
            root_names_reduces = copy(root_names)
            root_names_reduces.remove(root_name)

            P = base_ring_step[root_names_reduces[0]]
            p = P(monic_polynomial.monomial_coefficients())
            q, _ = p.quo_rem(P.gen() - first_root)

            verbose("Invoking recursion with: %s" % (q,))

            SplittingAlgebra.__init__(self, q, root_names_reduces, warning=False)

            splitting_roots = base_ring_step._splitting_roots + self._splitting_roots
            coefficients_list = base_ring_step._coefficients_list + self._coefficients_list

            verbose("Adding roots: %s" % (splitting_roots))

            self._splitting_roots = splitting_roots
            self._coefficients_list = coefficients_list
        else:
            PolynomialQuotientRing_domain.__init__(self, P, p, root_name)

            first_root = self.gen()
            self._splitting_roots.append(first_root)
            self._coefficients_list = [monic_polynomial.coefficients(sparse=False)]

            if not iterate:
                verbose("pre ring defined splitting_roots: %s" % (self._splitting_roots))
                return

            verbose("final ring defined splitting_roots: %s" % (self._splitting_roots))

        if deg == 2:
            coefficients = monic_polynomial.coefficients(sparse=False)
            lin_coeff = coefficients[1]
            self._splitting_roots.append(-lin_coeff - first_root)

        self._root_names = names
        self._splitting_roots = [self(root) for root in self._splitting_roots]
        verbose("splitting_roots: %s embedded" % (self._splitting_roots))

        # --------------------------------------------------------------------
        # try to calculate inverses of the roots. This is possible if the original polynomial
        # has an invertible constant term. For example let cf = [-w, v,-u, 1] that is
        # p = h^3 -u*h^2 + v*h -w, than u = x + y + z, v = x*y + x*z + y*z, w = x*y*z. If
        # w is invertible then 1/x = (v -(u-x)*x)/w, 1/y = (v -(u-y)*y)/w, 1/z = (v -(u-z)*z)/w
        # -----------------------------------------------------------------
        # first find the polynomial with invertible constant coefficient
        # -----------------------------------------------------------------
        cf0_inv = None
        for cf in self._coefficients_list:
            cf0 = cf[0]
            try:
                cf0_inv = ~(cf[0])
                cf0_inv = self(cf0_inv)
                verbose("invertible coefficient: %s found" % (cf0_inv))
                break
            except NotImplementedError:
                verbose("constant coefficient: %s not invertibe" % (cf0))

        # ------------------------------------------------------------------
        # assuming that cf splits into linear factors over self
        # and the _splitting_roots are its roots we can calculate inverses
        # ------------------------------------------------------------------
        if cf0_inv is not None:
            deg_cf = len(cf)-1
            pf = P(cf)
            for root in self._splitting_roots:
                check = self(pf)
                if not check.is_zero():
                    continue
                root_inv = self.one()
                for pos in range(deg_cf-1):
                    root_inv = (-1)**(pos+1) * cf[deg_cf-pos-1] - root_inv * root
                verbose("inverse %s of root %s" % (root_inv, root))
                root_inv = (-1)**(deg_cf) * cf0_inv * root_inv
                self._invertible_elements.update({root:root_inv})
                verbose("adding inverse %s of root %s" % (root_inv, root))
            invert_items = list(self._invertible_elements.items())
            for k, v in invert_items:
                self._invertible_elements.update({v: k})
        return

    ########################################################################
    # ----------------------------------------------------------------------
    # overloaded inherited methods
    # ----------------------------------------------------------------------
    ########################################################################
    def __reduce__(self):
        r"""
        Used in pickling.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<t, u, v, w > = LaurentPolynomialRing(ZZ); x = polygen(L)
            sage: S = SplittingAlgebra(x^4 -t*x^3 - u*x^2 - v*x + w, ('X', 'Y', 'Z'), warning=False)
            sage: S.__reduce__()
            (<class 'sage.algebras.splitting_algebra.SplittingAlgebra_with_category'>,
            (x^4 - t*x^3 - u*x^2 - v*x + w, ('X', 'Y', 'Z'), True, False))
            sage: S.base_ring().__reduce__()
            (<class 'sage.algebras.splitting_algebra.SplittingAlgebra_with_category'>,
            (Y^3 + (X - t)*Y^2 + (X^2 - t*X - u)*Y + X^3 - t*X^2 - u*X - v,
            ('Y', 'Z'),
            False,
            False))

            sage: TestSuite(S).run()
        """
        defining_polynomial = self.defining_polynomial()
        definig_coefficients = self._coefficients_list[0]
        if defining_polynomial.coefficients(sparse=False) != definig_coefficients:
            # case of factorization algebra (intermediate construction step)
            par_pol = self.cover_ring()
            defining_polynomial = par_pol(definig_coefficients)
        return self.__class__, (defining_polynomial, self._root_names, self._iterate, False)

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v> = PolynomialRing(ZZ)
            sage: t = polygen(L)
            sage: Spl.<S, T> = SplittingAlgebra(t^3 - (u^2-v)*t^2 + (v+u)*t - 1)
            sage: Spl._repr_()
            'Splitting Algebra of x^3 + (-u^2 + v)*x^2 + (u + v)*x - 1
             with roots [S, T, -T - S + u^2 - v]
             over Multivariate Polynomial Ring in u, v over Integer Ring'
            sage: Spl.base_ring()    # indirect doctest
            Factorization Algebra of x^3 + (-u^2 + v)*x^2 + (u + v)*x - 1
             with roots [S] over Multivariate Polynomial Ring in u, v over Integer Ring
        """
        if self.is_completely_split():
            return ('Splitting Algebra of %s with roots %s over %s'
                    % (self.defining_polynomial(), self.splitting_roots(), self.scalar_base_ring()))
        else:
            return ('Factorization Algebra of %s with roots %s over %s'
                    % (self.defining_polynomial(), self.splitting_roots(), self.scalar_base_ring()))

    def _first_ngens(self, n):
        r"""
        Used by the preparser for ``R.<x> = ...``.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v> = PolynomialRing(ZZ)
            sage: t = polygen(L)
            sage: S.<X, Y> = SplittingAlgebra(t^3 - (u^2-v)*t^2 + (v+u)*t - 1)  # indirect doctest
            sage: X.parent()
            Splitting Algebra of x^3 + (-u^2 + v)*x^2 + (u + v)*x - 1
             with roots [X, Y, -Y - X + u^2 - v]
             over Multivariate Polynomial Ring in u, v over Integer Ring
            sage: S._first_ngens(4)
            (X, Y, u, v)
        """
        srts = self.splitting_roots()
        k = len(srts)-1
        gens = srts[:k] + list(self.scalar_base_ring().gens())
        return tuple(gens[:n])

    def _element_constructor_(self, x):
        r"""
        Make sure ``x`` is a valid member of ``self``, and return the constructed element.

        TESTS::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v, w> = LaurentPolynomialRing(ZZ); x = polygen(L)
            sage: S.<X, Y> = SplittingAlgebra(x^3 - u*x^2 + v*x - w)
            sage: S(u + v)
            u + v
            sage: S(X*Y + X)
            X*Y + X
            sage: TestSuite(S).run()                   # indirect doctest
        """
        if isinstance(x, SplittingAlgebraElement):
            # coercion from covering fixes pickling problems
            return self(x.lift())
        return super()._element_constructor_(x)

    def hom(self, im_gens, codomain=None, check=True, base_map=None):
        r"""
        This version keeps track with the special recursive structure
        of :class:`SplittingAlgebra`

        Type ``Ring.hom?`` to see the general documentation of this method.
        Here you see just special examples for the current class.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v, w> = LaurentPolynomialRing(ZZ); x = polygen(L)
            sage: S = SplittingAlgebra(x^3 - u*x^2 + v*x - w, ('X', 'Y'))
            sage: P.<x, y, z> = PolynomialRing(ZZ)
            sage: F = FractionField(P)
            sage: im_gens = [F(g) for g in [y, x, x + y + z, x*y+x*z+y*z, x*y*z]]
            sage: f = S.hom(im_gens)
            sage: f(u), f(v), f(w)
            (x + y + z, x*y + x*z + y*z, x*y*z)
            sage: roots = S.splitting_roots(); roots
            [X, Y, -Y - X + u]
            sage: [f(r) for r in roots]
            [x, y, z]
        """
        base_ring = self.base_ring()

        if not isinstance(im_gens, (list, tuple)):
            im_gens = [im_gens]

        all_gens = self.gens_dict_recursive()
        if len(im_gens) != len(all_gens):
            return super().hom(im_gens, codomain=codomain, check=check, base_map=base_map)

        num_gens = len(self.gens())
        im_gens_start = [img for img in im_gens if im_gens.index(img) < num_gens]
        im_gens_end = [img for img in im_gens if im_gens.index(img) >= num_gens]

        if not im_gens_end:
            return super().hom(im_gens, codomain=codomain, check=check, base_map=base_map)

        verbose('base %s im_gens_end %s codomain %s check %s base_map %s' % (base_ring, im_gens_end, codomain, check, base_map))
        hom_on_base_recurs = base_ring.hom(im_gens_end, codomain=codomain, check=check, base_map=base_map)
        verbose('hom_on_base_recurs %s' % (hom_on_base_recurs))

        cover_ring = self.cover_ring()
        hom_from_cover = cover_ring.hom(im_gens_start, codomain=codomain, check=check, base_map=hom_on_base_recurs)
        lift = self.lifting_map()
        return hom_from_cover*lift

    ###################################################################
    # -----------------------------------------------------------------
    # local methods
    # -----------------------------------------------------------------
    ###################################################################

    ###################################################################
    # -----------------------------------------------------------------
    # global methods
    # -----------------------------------------------------------------
    ###################################################################

    def is_completely_split(self):
        r"""
        Return ``True`` if the defining polynomial of ``self`` splits into
        linear factors over ``self``.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v, w > = LaurentPolynomialRing(ZZ); x = polygen(L)
            sage: S.<a,b> = SplittingAlgebra(x^3 - u*x^2 + v*x - w)
            sage: S.is_completely_split()
            True
            sage: S.base_ring().is_completely_split()
            False
        """
        return len(self.splitting_roots()) >= self.defining_polynomial().degree()

    @cached_method
    def lifting_map(self):
        r"""
        Return a section map from ``self`` to the cover ring. It is implemented according
        to the same named method of :class:`~sage.rings.quotient_ring.QuotientRing_nc`.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: x = polygen(ZZ)
            sage: S = SplittingAlgebra(x^2+1, ('I',))
            sage: lift = S.lifting_map()
            sage: lift(5)
            5
            sage: r1, r2 =S.splitting_roots()
            sage: lift(r1)
            I
        """
        from sage.rings.morphism import RingMap_lift
        return RingMap_lift(self, self.cover_ring())

    def splitting_roots(self):
        r"""
        Return the roots of the split equation.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: x = polygen(ZZ)
            sage: S = SplittingAlgebra(x^2+1, ('I',))
            sage: S.splitting_roots()
            [I, -I]
        """
        return self._splitting_roots

    @cached_method
    def scalar_base_ring(self):
        r"""
        Return the ring of scalars of ``self`` (considered as an algebra).

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v, w > = LaurentPolynomialRing(ZZ)
            sage: x = polygen(L)
            sage: S = SplittingAlgebra(x^3 - u*x^2 + v*x - w, ('X', 'Y'))
            sage: S.base_ring()
            Factorization Algebra of x^3 - u*x^2 + v*x - w with roots [X]
             over Multivariate Laurent Polynomial Ring in u, v, w over Integer Ring
            sage: S.scalar_base_ring()
            Multivariate Laurent Polynomial Ring in u, v, w over Integer Ring
        """
        base_ring = self.base_ring()
        if isinstance(base_ring, SplittingAlgebra):
            if base_ring.is_completely_split():
                # another splitting algebra independent of self
                return base_ring
            else:
                return base_ring.scalar_base_ring()
        return base_ring

    @cached_method
    def defining_polynomial(self):
        r"""
        Return the defining polynomial of ``self``.

        EXAMPLES::

            sage: from sage.algebras.splitting_algebra import SplittingAlgebra
            sage: L.<u, v, w > = LaurentPolynomialRing(ZZ)
            sage: x = polygen(L)
            sage: S = SplittingAlgebra(x^3 - u*x^2 + v*x - w, ('X', 'Y'))
            sage: S.defining_polynomial()
            x^3 - u*x^2 + v*x - w
        """
        base_ring = self.base_ring()
        if isinstance(base_ring, SplittingAlgebra):
            if base_ring.is_completely_split():
                # another splitting algebra independent of self
                return self._defining_polynomial
            else:
                return base_ring.defining_polynomial()
        return self._defining_polynomial


# --------------------------------------------------------------------------------------------
# ============================================================================================
# Utility function to create the roots of a polynomial in an appropriate extension ring
# ============================================================================================
# --------------------------------------------------------------------------------------------

def solve_with_extension(monic_polynomial, root_names=None, var='x', flatten=False, warning=True):
    r"""
    Return all roots of a monic polynomial in its base ring or in an appropriate
    extension ring, as far as possible.

    INPUT:

    - ``monic_polynomial`` -- the monic polynomial whose roots should be created
    - ``root_names`` -- names for the indeterminates needed to define the
      splitting algebra of the ``monic_polynomial`` (if necessary and possible)
    - ``var`` -- (default: ``'x'``) for the indeterminate needed to define the
      splitting field of the ``monic_polynomial`` (if necessary and possible)
    - ``flatten`` -- boolean (default: ``True``); if ``True`` the roots will
      not be given as a list of pairs ``(root, multiplicity)`` but as a list of
      roots repeated according to their multiplicity
    - ``warning`` -- boolean (default: ``True``); can be used (by setting to
      ``False``) to suppress a warning which will be thrown whenever it cannot
      be checked that the Galois group of ``monic_polynomial`` is maximal

    OUTPUT:

    List of tuples ``(root, multiplicity)`` respectively list of roots repeated
    according to their multiplicity if option ``flatten`` is ``True``.

    EXAMPLES::

        sage: from sage.algebras.splitting_algebra import solve_with_extension
        sage: t = polygen(ZZ)
        sage: p = t^2 -2*t +1
        sage: solve_with_extension(p, flatten=True )
        [1, 1]
        sage: solve_with_extension(p)
        [(1, 2)]

        sage: cp5 = cyclotomic_polynomial(5, var='T').change_ring(UniversalCyclotomicField())
        sage: solve_with_extension(cp5)
        [(E(5), 1), (E(5)^4, 1), (E(5)^2, 1), (E(5)^3, 1)]
        sage: _[0][0].parent()
        Universal Cyclotomic Field
    """
    def create_roots(monic_polynomial, warning=True):
        r"""
        This internal function creates all roots of a polynomial in an
        appropriate extension ring assuming that none of the roots is
        contained its base ring.

        It first tries to create the splitting field of the given polynomial.
        If this is not faithful the splitting algebra will be created.

        INPUT:

        - ``monic_polynomial`` -- the monic polynomial whose roots should
          be created
        - ``warning`` -- boolean (default: ``True``); can be used (by setting
          to ``False``) to suppress a warning which will be thrown whenever it
          cannot be checked that the Galois group of ``monic_polynomial`` is
          maximal
        """
        parent = monic_polynomial.parent()
        base_ring = parent.base_ring()

        try:
            ext_field, embed = monic_polynomial.splitting_field(var, map=True)

            if embed.domain() != base_ring:
                # in this case the SplittingAlgebra is preferred
                raise NotImplementedError

            # -------------------------------------------------------------------------------------
            # in some cases the embedding of the base_ring in ext_field can not be obtained
            # as coercion
            # -------------------------------------------------------------------------------------
            reset_coercion = False
            from sage.rings.number_field.number_field import NumberField_generic
            if isinstance(base_ring, NumberField_generic):
                reset_coercion = True
            elif base_ring.is_finite() and not base_ring.is_prime_field():
                reset_coercion = True
            if reset_coercion:
                ext_field._unset_coercions_used()
                ext_field.register_coercion(embed)
                ext_field.register_conversion(embed)

            verbose("splitting field %s defined" % (ext_field))
            pol_emb = monic_polynomial.change_ring(ext_field)
            roots = pol_emb.roots()
        except NotImplementedError:
            ext_ring = SplittingAlgebra(monic_polynomial, name_list, warning=warning)
            verbose("splitting algebra %s defined" % (ext_ring))
            roots = [(r, 1) for r in ext_ring.splitting_roots()]
        return roots

    deg_pol = monic_polynomial.degree()
    if not root_names:
        from sage.structure.category_object import normalize_names
        root_names = normalize_names(deg_pol-1, 'r')
    name_list = list(root_names)
    root_list = []
    try:
        root_list = monic_polynomial.roots()
    except (TypeError, ValueError, NotImplementedError):
        pass

    if not root_list:
        # --------------------------------------------------------------
        # no roots found: find roots in an appropriate extension ring
        # --------------------------------------------------------------
        verbose("no roots in base_ring")
        if len(name_list) > deg_pol - 1:
            name_list = [name_list[i] for i in range(deg_pol - 1)]
        roots = create_roots(monic_polynomial, warning=warning)

    else:
        # ------------------------------------------------------------------
        # root calculation was possible but maybe some more roots in
        # an appropriate extension ring can be constructed.
        # ------------------------------------------------------------------
        num_roots = sum(m for r, m in root_list)
        if num_roots < deg_pol:
            h = monic_polynomial.variables()[0]
            divisor = monic_polynomial.base_ring().one()
            for r, m in root_list:
                divisor *= (h - r)**m
            q, r = monic_polynomial.quo_rem(divisor)
            if len(name_list) > deg_pol - num_roots - 1:
                name_list = [name_list[i] for i in range(deg_pol - num_roots - 1)]
            verbose("%d root found in base ring, now solving %s" % (num_roots,q))
            missing_roots = create_roots(q, warning=True)
            roots = root_list + missing_roots
        else:
            roots = root_list
            verbose("all roots in base ring")

    if flatten:
        from sage.misc.flatten import flatten
        return flatten([[rt]*m for rt, m in roots])
    return roots



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
"""
Binary recurrence sequences

This class implements several methods relating to general linear binary
recurrence sequences, including a sieve to find perfect powers in integral
linear binary recurrence sequences.

EXAMPLES::

    sage: R = BinaryRecurrenceSequence(1,1)        #the Fibonacci sequence
    sage: R(137)        #the 137th term of the Fibonacci sequence
    19134702400093278081449423917
    sage: R(137) == fibonacci(137)                                                      # needs sage.libs.pari
    True
    sage: [R(i) % 4 for i in range(12)]
    [0, 1, 1, 2, 3, 1, 0, 1, 1, 2, 3, 1]
    sage: R.period(4)        #the period of the fibonacci sequence modulo 4
    6
    sage: R.pthpowers(2, 10**10)        # long time (7 seconds) -- in fact these are all squares, c.f. [BMS06]
    [0, 1, 2, 12]

    sage: S = BinaryRecurrenceSequence(8,1)  #a Lucas sequence
    sage: S.period(73)
    148
    sage: S(5) % 73 == S(5 +148) %73
    True
    sage: S.pthpowers(3, 10**10)    # long time (3 seconds) -- provably finds the indices of all 3rd powers less than 10^10
    [0, 1, 2]

    sage: T = BinaryRecurrenceSequence(2,0,1,2)
    sage: [T(i) for i in range(10)]
    [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]
    sage: T.is_degenerate()
    True
    sage: T.is_geometric()
    True
    sage: T.pthpowers(7, 10**30)                                                        # needs sage.symbolic
    Traceback (most recent call last):
    ...
    ValueError: the degenerate binary recurrence sequence is geometric or quasigeometric
    and has many pth powers


AUTHORS:

- Isabel Vogt (2013): initial version

See [SV2013]_, [BMS2006]_, and [SS1983]_.
"""

# ****************************************************************************
#       Copyright (C) 2013 Isabel Vogt <ivogt161@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


from sage.structure.sage_object import SageObject
from sage.rings.finite_rings.integer_mod_ring import Integers
from sage.rings.finite_rings.finite_field_constructor import GF
from sage.rings.integer import Integer
from sage.arith.functions import lcm
from sage.arith.misc import is_prime, next_prime, next_prime_power, legendre_symbol
from sage.misc.lazy_import import lazy_import
from sage.misc.functional import sqrt

lazy_import("sage.functions.log", "log")
lazy_import("sage.rings.number_field.number_field", "QuadraticField")
lazy_import("sage.matrix.constructor", "matrix")
lazy_import("sage.modules.free_module_element", "vector")


class BinaryRecurrenceSequence(SageObject):
    """
    Create a linear binary recurrence sequence defined by initial conditions
    `u_0` and `u_1` and recurrence relation `u_{n+2} = b*u_{n+1}+c*u_n`.

    INPUT:

    - ``b`` -- integer; (partially determining the recurrence relation)

    - ``c`` -- integer; (partially determining the recurrence relation)

    - ``u0`` -- integer; (the `0`-th term of the binary recurrence sequence)

    - ``u1`` -- integer; (the `1`-st term of the binary recurrence sequence)

    OUTPUT: an integral linear binary recurrence sequence defined by `u_0`,
    `u_1`, and `u_{n+2} = b u_{n+1}+c u_n`

    .. SEEALSO::

        :func:`fibonacci`, :func:`lucas_number1`, :func:`lucas_number2`

    EXAMPLES::

        sage: R = BinaryRecurrenceSequence(3,3,2,1)
        sage: R
        Binary recurrence sequence defined by: u_n = 3 * u_{n-1} + 3 * u_{n-2};
        With initial conditions: u_0 = 2, and u_1 = 1
    """

    def __init__(self, b, c, u0=0, u1=1):
        """
        See :class:`BinaryRecurrenceSequence` for full documentation.

        EXAMPLES::

            sage: R = BinaryRecurrenceSequence(3,3,2,1)
            sage: R
            Binary recurrence sequence defined by: u_n = 3 * u_{n-1} + 3 * u_{n-2};
            With initial conditions: u_0 = 2, and u_1 = 1

            sage: R = BinaryRecurrenceSequence(1,1)
            sage: loads(R.dumps()) == R
            True
        """
        self.b = b
        self.c = c
        self.u0 = u0
        self.u1 = u1
        self._period_dict = {}  # dictionary to cache the period of a sequence for future lookup
        self._PGoodness = {}  # dictionary to cache primes that are "good" by some prime power
        self._ell = 1  # variable that keeps track of the last prime power to be used as a goodness

    def __repr__(self) -> str:
        """
        Give string representation of the class.

        EXAMPLES::

            sage: R = BinaryRecurrenceSequence(3,3,2,1)
            sage: R
            Binary recurrence sequence defined by: u_n = 3 * u_{n-1} + 3 * u_{n-2};
            With initial conditions: u_0 = 2, and u_1 = 1
        """
        return 'Binary recurrence sequence defined by: u_n = ' + str(self.b) + ' * u_{n-1} + ' + str(self.c) + ' * u_{n-2};\nWith initial conditions: u_0 = ' + str(self.u0) + ', and u_1 = ' + str(self.u1)

    def __eq__(self, other) -> bool:
        """
        Compare two binary recurrence sequences.

        EXAMPLES::

            sage: R = BinaryRecurrenceSequence(3,3,2,1)
            sage: S = BinaryRecurrenceSequence(3,3,2,1)
            sage: R == S
            True

            sage: T = BinaryRecurrenceSequence(3,3,2,2)
            sage: R == T
            False
        """
        return (self.u0 == other.u0) and (self.u1 == other.u1) and (self.b == other.b) and (self.c == other.c)

    def __call__(self, n, modulus=0):
        """
        Give the `n`-th term of a binary recurrence sequence, possibly mod some modulus.

        INPUT:

        - ``n`` -- integer; the index of the term in the binary recurrence sequence

        - ``modulus`` -- a natural number (optional --  default value is 0)

        OUTPUT:

        - An integer (the `n`-th term of the binary recurrence sequence modulo ``modulus``)

        EXAMPLES::

            sage: R = BinaryRecurrenceSequence(3,3,2,1)
            sage: R(2)
            9
            sage: R(101)
            16158686318788579168659644539538474790082623100896663971001
            sage: R(101,12)
            9
            sage: R(101)%12
            9
        """
        R = Integers(modulus)
        F = matrix(R, [[0, 1], [self.c, self.b]])
        # F*[u_{n}, u_{n+1}]^T = [u_{n+1}, u_{n+2}]^T (T indicates transpose).
        v = vector(R, [self.u0, self.u1])
        return list(F**n * v)[0]

    def is_degenerate(self) -> bool:
        """
        Decide whether the binary recurrence sequence is degenerate.

        Let `\\alpha` and `\\beta` denote the roots of the characteristic polynomial
        `p(x) = x^2-bx -c`.  Let `a = u_1-u_0\\beta/(\\beta - \\alpha)` and
        `b = u_1-u_0\\alpha/(\\beta - \\alpha)`.  The sequence is, thus, given by
        `u_n = a \\alpha^n - b\\beta^n`.  Then we say that the sequence is nondegenerate
        if and only if `a*b*\\alpha*\\beta \\neq 0` and `\\alpha/\\beta` is not a
        root of unity.

        More concretely, there are 4 classes of degeneracy, that can all be formulated
        in terms of the matrix `F = [[0,1], [c, b]]`.

        - `F` is singular -- this corresponds to ``c`` = 0, and thus
          `\\alpha*\\beta = 0`. This sequence is geometric after term ``u0``
          and so we call it ``quasigeometric``

        - `v = [[u_0], [u_1]]` is an eigenvector of `F` -- this corresponds to
          a ``geometric`` sequence with `a*b = 0`

        - `F` is nondiagonalizable -- this corresponds to `\\alpha = \\beta`.
          This sequence will be the point-wise product of an arithmetic and
          geometric sequence.

        - `F^k` is scalar, for some `k>1` -- this corresponds to
          `\\alpha/\\beta` a `k` th root of unity. This sequence is a union of
          several geometric sequences, and so we again call it ``quasigeometric``.

        EXAMPLES::

            sage: S = BinaryRecurrenceSequence(0,1)
            sage: S.is_degenerate()
            True
            sage: S.is_geometric()
            False
            sage: S.is_quasigeometric()
            True

            sage: R = BinaryRecurrenceSequence(3,-2)
            sage: R.is_degenerate()
            False

            sage: T = BinaryRecurrenceSequence(2,-1)
            sage: T.is_degenerate()
            True
            sage: T.is_arithmetic()
            True
        """
        D = self.b**2 + 4 * self.c
        if D != 0:
            if D.is_square():
                A = sqrt(D)
            else:
                A = QuadraticField(D, 'x').gen()

            aa = (self.u1 - self.u0 * (self.b + A)/2)/(A)     # called `a` in Docstring
            bb = (self.u1 - self.u0 * (self.b - A)/2)/(A)     # called `b` in Docstring

            # (b+A)/2 is called alpha in Docstring, (b-A)/2 is called beta in Docstring

            if self.b != A:
                if ((self.b+A)/(self.b-A))**6 == 1:
                    return True
            else:
                return True

            return aa*bb*(self.b + A)*(self.b - A) == 0

        return True

    def is_geometric(self) -> bool:
        """
        Decide whether the binary recurrence sequence is geometric - ie a geometric sequence.

        This is a subcase of a degenerate binary recurrence sequence, for which `ab=0`, i.e.
        `u_{n}/u_{n-1}=r` for some value of `r`.

        See :meth:`is_degenerate` for a description of
        degeneracy and definitions of `a` and `b`.

        EXAMPLES::

            sage: S = BinaryRecurrenceSequence(2,0,1,2)
            sage: [S(i) for i in range(10)]
            [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]
            sage: S.is_geometric()
            True
        """
        # If [u_0, u_1]^T is an eigenvector for the incrementation matrix F = [[0,1],[c,b]], then the sequence
        # is geometric, ie we can write u_n = a*r^n for some a and r.

        # We decide if u0, u1, u2 = b*u1+c*u0 are in geometric progression by whether u1^2 = (b*u1+c*u0)*u0

        return (self.u1)**2 == (self.b*self.u1 + self.c*self.u0)*self.u0

    def is_quasigeometric(self) -> bool:
        """
        Decide whether the binary recurrence sequence is degenerate and similar to a geometric sequence,
        i.e. the union of multiple geometric sequences, or geometric after term ``u0``.

        If `\\alpha/\\beta` is a `k` th root of unity, where `k>1`, then necessarily `k = 2, 3, 4, 6`.
        Then `F = [[0,1],[c,b]` is diagonalizable, and `F^k = [[\\alpha^k, 0], [0,\\beta^k]]` is a diagonal
        matrix.  Thus for all values of `j` mod `k`, the `j` mod `k` terms of `u_n` form a geometric
        series.

        If `\\alpha` or `\\beta` is zero, this implies that `c=0`.  This is the case when `F` is
        singular.  In this case, `u_1, u_2, u_3, ...` is geometric.

        EXAMPLES::

            sage: S = BinaryRecurrenceSequence(0,1)
            sage: [S(i) for i in range(10)]
            [0, 1, 0, 1, 0, 1, 0, 1, 0, 1]
            sage: S.is_quasigeometric()
            True

            sage: R = BinaryRecurrenceSequence(3,0)
            sage: [R(i) for i in range(10)]
            [0, 1, 3, 9, 27, 81, 243, 729, 2187, 6561]
            sage: R.is_quasigeometric()
            True
        """
        # First test if F is singular... i.e. beta = 0
        if self.c == 0:
            return True

        # Otherwise test if alpha/beta is a root of unity that is not 1
        D = self.b**2 + 4 * self.c
        if D != 0:    # thus alpha/beta != 1
            if D.is_square():
                A = sqrt(D)
            else:
                A = QuadraticField(D, 'x').gen()
            if ((self.b+A)/(self.b-A))**6 == 1:
                return True

        return False

    def is_arithmetic(self) -> bool:
        """
        Decide whether the sequence is degenerate and an arithmetic sequence.

        The sequence is arithmetic if and only if `u_1 - u_0 = u_2 - u_1 = u_3 - u_2`.

        This corresponds to the matrix `F = [[0,1],[c,b]]` being nondiagonalizable
        and `\\alpha/\\beta = 1`.

        EXAMPLES::

            sage: S = BinaryRecurrenceSequence(2,-1)
            sage: [S(i) for i in range(10)]
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            sage: S.is_arithmetic()
            True
        """
        return self(1) - self(0) == self(2) - self(1) == self(3) - self(2)

    def period(self, m, *, eventual=False):
        """
        Return the period of the binary recurrence sequence modulo
        an integer ``m``.

        If `n_1` is congruent to `n_2` modulo ``period(m)``, then `u_{n_1}` is
        is congruent to `u_{n_2}` modulo ``m``.

        INPUT:

        - ``m`` -- integer; modulo which the period of the recurrence relation is calculated

        - ``eventual`` -- boolean (default: ``False``); if ``True``, allow the
          sequence to be eventually periodic, rather than requiring it to be
          purely periodic. So `n_1` might not be congruent to `n_2` modulo
          ``period(m)`` unless `n_1` and `n_2` are large.

        OUTPUT: integer (the period of the sequence modulo m)

        EXAMPLES:

        If `p = \\pm 1 \\mod 5`, then the period of the Fibonacci sequence
        mod `p` is `p-1` (c.f. Lemma 3.3 of [BMS2006]_).

        ::

            sage: R = BinaryRecurrenceSequence(1,1)
            sage: R.period(31)
            30

            sage: [R(i) % 4 for i in range(12)]
            [0, 1, 1, 2, 3, 1, 0, 1, 1, 2, 3, 1]
            sage: R.period(4)
            6

        This function works for degenerate sequences as well.

        ::

            sage: S = BinaryRecurrenceSequence(2,0,1,2)
            sage: S.is_degenerate()
            True
            sage: S.is_geometric()
            True
            sage: [S(i) % 17 for i in range(16)]
            [1, 2, 4, 8, 16, 15, 13, 9, 1, 2, 4, 8, 16, 15, 13, 9]
            sage: S.period(17)
            8

        Letting ``eventual`` be ``True`` allows us to find the period of a
        sequence that is not purely periodic. ::

            sage: T = BinaryRecurrenceSequence(5,12,u0=0,u1=1)
            sage: [T(n) % 10 for n in range(20)]
            [0, 1, 5, 7, 5, 9, 5, 3, 5, 1, 5, 7, 5, 9, 5, 3, 5, 1, 5, 7]
            sage: T.period(10)
            Traceback (most recent call last):
            ...
            ValueError: Binary recurrence sequence modulo 10 is not a purely
            periodic sequence.
            sage: T.period(10,eventual=True)
            8

        .. NOTE:: The answer is cached.

        TESTS:

        Verify that :issue:`38112` is fixed::

            sage: T = BinaryRecurrenceSequence(3,2,u0=0,u1=1)
            sage: [T(n) % 4 for n in range(5)]
            [0, 1, 3, 3, 3]
            sage: T.period(4)
            Traceback (most recent call last):
            ...
            ValueError: Binary recurrence sequence modulo 4 is not a purely
            periodic sequence.
            sage: T.period(4,eventual=True)
            1
        """

        # If we have already computed the period mod m, then we return the stored value.

        if m in self._period_dict:
            return self._period_dict[m]

        else:
            R = Integers(m)
            A = matrix(R, [[0, 1], [self.c, self.b]])
            w = vector(R, [self.u0, self.u1])
            Fac = list(m.factor())
            Periods = {}

            if eventual is True:
                # There are only m^2 possible pairs modulo m. Since the numbering
                # of the sequence starts at 0, this implies that the term numbered
                # m^2 must be in periodic part of the sequence. Hence, the
                # sequence starting with the terms numbered m^2 and m^2 + 1 must
                # be purely periodic.
                an = (A**(m**2)) * w
                return BinaryRecurrenceSequence(self.b, self.c,
                    an[0], an[1]).period(m, eventual=False)

            # To compute the period mod m, we compute the least integer n such that A^n*w == w.  This necessarily
            # divides the order of A as a matrix in GL_2(Z/mZ).

            # We compute the period modulo all distinct prime powers dividing m, and combine via the lcm.
            # To compute the period mod p^e, we first compute the order mod p.  Then the period mod p^e
            # must divide p^{4e-4}*period(p), as the subgroup of matrices mod p^e, which reduce to
            # the identity mod p is of order (p^{e-1})^4.  So we compute the period mod p^e by successively
            # multiplying the period mod p by powers of p.

            for p, e in Fac:
                # first compute the period mod p
                if p in self._period_dict:
                    perp = self._period_dict[p]
                else:
                    F = A.change_ring(GF(p))
                    v = w.change_ring(GF(p))
                    FF = F**(p-1)
                    p1fac = list((p-1).factor())

                    # The order of any matrix in GL_2(F_p) either divides p(p-1) or (p-1)(p+1).
                    # The order divides p-1 if it is diagonalizable.  In any case, det(F^(p-1))=1,
                    # so if tr(F^(p-1)) = 2, then it must be triangular of the form [[1,a],[0,1]].
                    # The order of the subgroup of matrices of this form is p, so the order must divide
                    # p(p-1) -- in fact it must be a multiple of p.  If this is not the case, then the
                    # order divides (p-1)(p+1).  As the period divides the order of the matrix in GL_2(F_p),
                    # these conditions hold for the period as well.

                    # check if the order divides (p-1)
                    if FF*v == v:
                        M = p-1
                        Mfac = p1fac

                    # check if the trace is 2, then the order is a multiple of p dividing p*(p-1)
                    elif FF.trace() == 2:
                        M = p-1
                        Mfac = p1fac
                        F = F**p        # replace F by F^p as now we only need to determine the factor dividing (p-1)

                    # otherwise it will divide (p+1)(p-1)
                    else:
                        M = (p+1)*(p-1)
                        p2fac = list((p+1).factor())        # factor the (p+1) and (p-1) terms separately and then combine for speed
                        Mfac_dic = {}
                        for i0, i1 in list(p1fac + p2fac):
                            if i0 not in Mfac_dic:
                                Mfac_dic[i0] = i1
                            else:
                                Mfac_dic[i0] += i1
                        Mfac = list(Mfac_dic.items())

                    # Now use a fast order algorithm to compute the period.  We know that the period divides
                    # M = i_1*i_2*...*i_l where the i_j denote not necessarily distinct prime factors.  As
                    # F^M*v == v, for each i_j, if F^(M/i_j)*v == v, then the period divides (M/i_j).  After
                    # all factors have been iterated over, the result is the period mod p.

                    Mfac = list(Mfac)

                    # expand the list of prime factors so every factor is with multiplicity 1
                    C = [i0 for i0, i1 in Mfac for _ in range(i1)]

                    Mfac = C
                    n = M
                    for ii in Mfac:
                        b = n // ii
                        if F**b * v == v:
                            n = b
                    perp = n

                # Now compute the period mod p^e by stepping up by multiples of p
                F = A.change_ring(Integers(p**e))
                v = w.change_ring(Integers(p**e))
                FF = F**perp
                if FF*v == v:
                    perpe = perp
                else:
                    tries = 0
                    while True:
                        tries += 1
                        FF = FF**p
                        if FF*v == v:
                            perpe = perp*p**tries
                            break
                        if tries > e:
                            raise ValueError("Binary recurrence sequence " +
                                             f"modulo {m} is not a purely " +
                                             "periodic sequence.")
                Periods[p] = perpe

            # take the lcm of the periods mod all distinct primes dividing m
            period = lcm(Periods.values())

            self._period_dict[m] = period        # cache the period mod m
            return period

    def pthpowers(self, p, Bound):
        """
        Find the indices of proveably all `p`-th powers in the recurrence sequence
        bounded by Bound.

        Let `u_n` be a binary recurrence sequence.  A ``p`` th power in `u_n`
        is a solution to `u_n = y^p` for some integer `y`.  There are only
        finitely many ``p`` th powers in any recurrence sequence [SS1983]_.

        INPUT:

        - ``p`` -- a rational prime integer (the fixed p in `u_n = y^p`)

        - ``Bound`` -- a natural number (the maximum index `n` in `u_n = y^p` that is checked)

        OUTPUT:

        A list of the indices of all ``p`` th powers less bounded by
        ``Bound``. If the sequence is degenerate and there are many
        ``p`` th powers, raises :exc:`ValueError`.

        EXAMPLES::

            sage: R = BinaryRecurrenceSequence(1,1)        #the Fibonacci sequence
            sage: R.pthpowers(2, 10**10)        # long time (7 seconds) -- in fact these are all squares, c.f. [BMS2006]_
            [0, 1, 2, 12]

            sage: S = BinaryRecurrenceSequence(8,1) #a Lucas sequence
            sage: S.pthpowers(3,10**10)    # long time (3 seconds) -- provably finds the indices of all 3rd powers less than 10^10
            [0, 1, 2]

            sage: Q = BinaryRecurrenceSequence(3,3,2,1)
            sage: Q.pthpowers(11,10**10)          # long time (7.5 seconds)
            [1]

        If the sequence is degenerate, and there are no ``p`` th powers, returns `[]`.  Otherwise, if
        there are many ``p`` th powers, raises :exc:`ValueError`.

        ::

            sage: T = BinaryRecurrenceSequence(2,0,1,2)
            sage: T.is_degenerate()
            True
            sage: T.is_geometric()
            True
            sage: T.pthpowers(7, 10**30)                                                # needs sage.symbolic
            Traceback (most recent call last):
            ...
            ValueError: the degenerate binary recurrence sequence is geometric or
            quasigeometric and has many pth powers

            sage: L = BinaryRecurrenceSequence(4,0,2,2)
            sage: [L(i).factor() for i in range(10)]
            [2, 2, 2^3, 2^5, 2^7, 2^9, 2^11, 2^13, 2^15, 2^17]
            sage: L.is_quasigeometric()
            True
            sage: L.pthpowers(2, 10**30)                                                # needs sage.symbolic
            []

        .. NOTE::

            This function is primarily optimized in the range where
            ``Bound`` is much larger than ``p``.
        """
        # Thanks to Jesse Silliman for helpful conversations!

        # Reset the dictionary of good primes, as this depends on p
        self._PGoodness = {}
        # Starting lower bound on good primes
        self._ell = 1

        # If the sequence is geometric, then the `n`th term is `a*r^n`.  Thus the
        # property of being a ``p`` th power is periodic mod ``p``.  So there are either
        # no ``p`` th powers if there are none in the first ``p`` terms, or many if there
        # is at least one in the first ``p`` terms.

        if self.is_geometric() or self.is_quasigeometric():
            no_powers = True
            for i in range(1, 6*p+1):
                if _is_p_power(self(i), p):
                    no_powers = False
                    break
            if no_powers:
                if _is_p_power(self.u0, p):
                    return [0]
                return []
            else:
                raise ValueError("the degenerate binary recurrence sequence is geometric or quasigeometric and has many pth powers")

        # If the sequence is degenerate without being geometric or quasigeometric, there
        # may be many ``p`` th powers or no ``p`` th powers.

        elif (self.b**2+4*self.c) == 0:

            # This is the case if the matrix F is not diagonalizable, ie b^2 +4c = 0, and alpha/beta = 1.

            alpha = self.b/2

            # In this case, u_n = u_0*alpha^n + (u_1 - u_0*alpha)*n*alpha^(n-1) = alpha^(n-1)*(u_0 +n*(u_1 - u_0*alpha)),
            # that is, it is a geometric term (alpha^(n-1)) times an arithmetic term (u_0 + n*(u_1-u_0*alpha)).

            # Look at classes n = k mod p, for k = 1,...,p.

            for k in range(1, p + 1):

                # The linear equation alpha^(k-1)*u_0 + (k+pm)*(alpha^(k-1)*u1 - u0*alpha^k)
                # must thus be a pth power.  This is a linear equation in m, namely, A + B*m, where

                A = (alpha**(k-1)*self.u0 + k*(alpha**(k-1)*self.u1 - self.u0*alpha**k))
                B = p*(alpha**(k-1)*self.u1 - self.u0*alpha**k)

                # This linear equation represents a pth power iff A is a pth power mod B.

                if _is_p_power_mod(A, p, B):
                    raise ValueError("the degenerate binary recurrence sequence has many pth powers")
            return []

        # We find ``p`` th powers using an elementary sieve.  Term `u_n` is a ``p`` th
        # power if and only if it is a ``p`` th power modulo every prime `\\ell`.  This condition
        # gives nontrivial information if ``p`` divides the order of the multiplicative group of
        # `\\Bold(F)_{\\ell}`, i.e. if `\\ell` is ` 1 \mod{p}`, as then only `1/p` terms are ``p`` th
        # powers modulo `\\ell``.

        # Thus, given such an `\\ell`, we get a set of necessary congruences for the index modulo the
        # the period of the sequence mod `\\ell`.  Then we intersect these congruences for many primes
        # to get a tight list modulo a growing modulus.  In order to keep this step manageable, we
        # only use primes `\\ell` that have particularly smooth periods.

        # Some congruences in the list will remain as the modulus grows.  If a congruence remains through
        # 7 rounds of increasing the modulus, then we check if this corresponds to a perfect power (if
        # it does, we add it to our list of indices corresponding to ``p`` th powers).  The rest of the congruences
        # are transient and grow with the modulus.  Once the smallest of these is greater than the bound,
        # the list of known indices corresponding to ``p`` th powers is complete.

        else:

            if Bound < 3 * p:

                powers = []
                ell = p + 1

                while not is_prime(ell):
                    ell += p

                F = GF(ell)
                a0 = F(self.u0)
                a1 = F(self.u1)  # a0 and a1 are variables for terms in sequence
                bf, cf = F(self.b), F(self.c)

                for n in range(Bound):  # n is the index of the a0

                    # Check whether a0 is a perfect power mod ell
                    if _is_p_power_mod(a0, p, ell):
                        # if a0 is a perfect power mod ell, check if nth term is ppower
                        if _is_p_power(self(n), p):
                            powers.append(n)

                    a0, a1 = a1, bf*a1 + cf*a0      # step up the variables

            else:

                powers = []      # documents the indices of the sequence that provably correspond to pth powers
                cong = [0]       # list of necessary congruences on the index for it to correspond to pth powers
                Possible_count = {}   # keeps track of the number of rounds a congruence lasts in cong

                # These parameters are involved in how we choose primes to increase the modulus
                qqold = 1        # we believe that we know complete information coming from primes good by qqold
                M1 = 1            # we have congruences modulo M1, this may not be the tightest list
                M2 = p            # we want to move to have congruences mod M2
                qq = 1            # the largest prime power divisor of M1 is qq

                # This loop ups the modulus.
                while True:

                    # Try to get good data mod M2

                    # patience of how long we should search for a "good prime"
                    patience = 0.01 * _estimated_time(lcm(M2, p * next_prime_power(qq)),
                                                      M1, len(cong), p)
                    tries = 0

                    # This loop uses primes to get a small set of congruences mod M2.
                    while True:

                        # only proceed if took less than patience time to find the next good prime
                        ell = _next_good_prime(p, self, qq, patience, qqold)
                        if ell:

                            # gather congruence data for the sequence mod ell, which will be mod period(ell) = modu
                            cong1, modu = _find_cong1(p, self, ell)

                            # makes a new list from cong that is now mod M = lcm(M1, modu) instead of M1
                            M = lcm(M1, modu)
                            CongNew = [k * M1 + i for k in range(M // M1)
                                       for i in cong]
                            cong = set(CongNew)

                            M1 = M

                            killed_something = False        # keeps track of when cong1 can rule out a congruence in cong

                            # CRT by hand to gain speed
                            for i in list(cong):
                                if i % modu not in cong1:     # congruence in cong is inconsistent with any in cong1
                                    cong.remove(i)            # remove that congruence
                                    killed_something = True

                            if M1 == M2:
                                if not killed_something:
                                    tries += 1
                                    if tries == 2:            # try twice to rule out congruences
                                        cong = list(cong)
                                        qqold = qq
                                        qq = next_prime_power(qq)
                                        M2 = lcm(M2, p * qq)
                                        break

                        else:
                            qq = next_prime_power(qq)
                            M2 = lcm(M2, p * qq)
                            cong = list(cong)
                            break

                    # Document how long each element of cong has been there
                    for i in cong:
                        if i in Possible_count:
                            Possible_count[i] += 1
                        else:
                            Possible_count[i] = 1

                    # Check how long each element has persisted, if it is for at least 7 cycles,
                    # then we check to see if it is actually a perfect power
                    for i, pci in Possible_count.items():
                        if pci == 7:
                            n = Integer(i)
                            if n < Bound:
                                if _is_p_power(self(n), p):
                                    powers.append(n)

                    # check for a contradiction
                    if len(cong) > len(powers):
                        if cong[len(powers)] > Bound:
                            break
                    elif M1 > Bound:
                        break

            return powers


def _prime_powers(N):
    r"""
    Find the prime powers dividing `N`.

    In other words, if `N = q_1^{e_1} q_2^{e_2} \cdots q_n^{e_n}`, it returns
    `[q_1^{e_1}, q_2^{e_2}, \ldots, q_n^{e_n}]`.

    INPUT:

    - ``N`` -- integer

    OUTPUT: list of the prime powers dividing N

    EXAMPLES::

        sage: sage.combinat.binary_recurrence_sequences._prime_powers(124656)
        [3, 16, 49, 53]

        sage: sage.combinat.binary_recurrence_sequences._prime_powers(65537)
        [65537]
    """
    return sorted(i**j for i, j in N.factor())


def _largest_ppower_divisor(N):
    """
    Find the largest prime power divisor of `N`.

    INPUT:

    - ``N`` -- integer

    OUTPUT: the largest prime power dividing `N`

    EXAMPLES::

        sage: sage.combinat.binary_recurrence_sequences._largest_ppower_divisor(124656)
        53
        sage: sage.combinat.binary_recurrence_sequences._largest_ppower_divisor(65537)
        65537
    """
    return _prime_powers(N)[-1]


def _goodness(n, R, p):
    """
    Return the goodness of `n` for the sequence `R` and the prime `p` -- that is the largest
    non-`p` prime power dividing ``period(n)``.

    INPUT:

    - ``n`` -- an integer

    - ``R`` -- an object in the class ``BinaryRecurrenceSequence``

    - ``p`` -- a rational prime

    OUTPUT:

    - An integer which is the "goodness" of ``n``, i.e. the largest non-``p`` prime power dividing ``period(n)``.

    EXAMPLES::

        sage: R = BinaryRecurrenceSequence(11,2)
        sage: sage.combinat.binary_recurrence_sequences._goodness(89,R,7)
        11

        sage: R = BinaryRecurrenceSequence(1,1)
        sage: sage.combinat.binary_recurrence_sequences._goodness(13,R,7)
        4
        sage: R.period(13)        #the period of R mod 13 is divisible by 7
        28
    """
    # The period of R mod ell
    K = R.period(n)
    return _largest_ppower_divisor(K // K.gcd(p))


def _next_good_prime(p, R, qq, patience, qqold):
    r"""
    Find the next prime `\ell` which is good by ``qq`` but not by ``qqold``, 1 mod ``p``, and for which
    ``b^2+4*c`` is a square mod `\ell`, for the sequence ``R`` if it is possible in runtime patience.

    INPUT:

    - ``p`` -- a prime

    - ``R`` -- an object in the class ``BinaryRecurrenceSequence``

    - ``qq`` -- a perfect power

    - ``patience`` -- a real number

    - ``qqold`` -- a perfect power less than or equal to ``qq``

    OUTPUT: a prime `\ell` such that `\ell` is 1 mod `p`, `b^2+4 c` is a
    square mod `\ell` and the period of `\ell` has ``goodness`` by ``qq`` but
    not ``qqold``, if patience has not be surpased; otherwise ``False``

    EXAMPLES::

        sage: # needs sage.libs.pari
        sage: R = BinaryRecurrenceSequence(1,1)
        sage: sage.combinat.binary_recurrence_sequences._next_good_prime(7,R,1,100,1)        #ran out of patience to search for good primes
        False
        sage: sage.combinat.binary_recurrence_sequences._next_good_prime(7,R,2,100,1)
        29
        sage: sage.combinat.binary_recurrence_sequences._next_good_prime(7,R,2,100,2)        #ran out of patience, as qqold == qq, so no primes work
        False
    """
    # We are looking for pth powers in R.
    # Our primes must be good by qq, but not qqold.
    # We only allow patience number of iterations to find a good prime.

    # The variable _ell for R keeps track of the last "good" prime returned
    # that was not found from the dictionary _PGoodness

    # First, we check to see if we have already computed the goodness of a prime that fits
    # our requirement of being good by qq but not by qqold.  This is stored in the _PGoodness
    # dictionary.

    # Then if we have, we return the smallest such prime and delete it from the list.  If not, we
    # search through patience number of primes R._ell to find one good by qq but not qqold.  If it is
    # not good by either qqold or qq, then we add this prime to R._PGoodness under its goodness.

    # Possible_Primes keeps track of possible primes satisfying our goodness requirements we might return
    # check to see if anything in R._PGoodness fits our goodness requirements
    Possible_Primes = [item[0] for j, item in R._PGoodness.items()
                       if qqold < j <= qq and item]

    # If we found good primes, we take the smallest
    if Possible_Primes:
        q = min(Possible_Primes)
        n = _goodness(q, R, p)
        del R._PGoodness[n][0]   # if we are going to use it, then we delete it from R._PGoodness
        return q

    # If nothing is already stored in R._PGoodness, we start (from where we left off at R._ell) checking
    # for good primes.  We only tolerate patience number of tries before giving up.
    else:
        i = 0
        while i < patience:
            i += 1
            R._ell = next_prime(R._ell)

            # we require that R._ell is 1 mod p, so that p divides the order of the multiplicative
            # group mod R._ell, so that not all elements of GF(R._ell) are pth powers.
            if R._ell % p == 1:

                # requiring that b^2 + 4c is a square in GF(R._ell) ensures that the period mod R._ell
                # divides R._ell - 1
                if legendre_symbol(R.b**2 + 4*R.c, R._ell) == 1:

                    N = _goodness(R._ell, R, p)

                    # proceed only if R._ell satisfies the goodness requirements
                    if qqold < N <= qq:
                        return R._ell

                    # if we do not use the prime, we store it in R._PGoodness
                    else:
                        if N in R._PGoodness:
                            R._PGoodness[N].append(R._ell)
                        else:
                            R._PGoodness[N] = [R._ell]

        return False


def _is_p_power_mod(a, p, N):
    """
    Determine if ``a`` is a ``p`` th power modulo ``N``.

    By the CRT, this is equivalent to the condition that ``a`` be a ``p`` th power mod all
    distinct prime powers dividing ``N``.  For each of these, we use the strong statement of
    Hensel's lemma to lift ``p`` th powers mod `q` or `q^2` or `q^3` to ``p`` th powers mod `q^e`.

    INPUT:

    - ``a`` -- integer

    - ``p`` -- a rational prime number

    - ``N`` -- positive integer

    OUTPUT: ``True`` if `a` is a `p`-th power modulo `N`; ``False`` otherwise

    EXAMPLES::

        sage: sage.combinat.binary_recurrence_sequences._is_p_power_mod(2**3,7,29)
        False
        sage: sage.combinat.binary_recurrence_sequences._is_p_power_mod(2**3,3,29)
        True
    """
    # By the chinese remainder theorem, we can answer this question by examining whether
    # a is a pth power mod q^e, for all distinct prime powers q^e dividing N.

    for q, e in N.factor():

        # If a = q^v*x, with

        v = a.valuation(q)

        # then if v>=e, a is congruent to 0 mod q^e and is thus a pth power trivially.

        if v >= e:
            continue

        # otherwise, it can only be a pth power if v is a multiple of p.
        if v % p:
            return False

        # in this cse it is a pth power if x is a pth power mod q^(e-v), so let x = aa,
        # and (e-v) = ee:

        aa = a / q**v
        ee = e - v

        # The above steps are equivalent to the statement that we may assume a and qq are
        # relatively prime, if we replace a with aa and e with ee.  Now we must determine when
        # aa is a pth power mod q^ee for (aa,q)=1.

        # If q != p, then by Hensel's lemma, we may lift a pth power mod q, to a pth power
        # mod q^2, etc.

        if q != p:

            # aa is necessarily a pth power mod q if p does not divide the order of the multiplicative
            # group mod q, ie if q is not 1 mod p.

            if q % p == 1:

                # otherwise aa if a pth power mod q iff aa^(q-1)/p == 1

                if GF(q)(aa)**((q - 1) / p) != 1:
                    return False

        # If q = p and ee = 1, then everything is a pth power p by Fermat's little theorem.

        elif ee > 1:

            # We use the strong statement of Hensel's lemma, which implies that if p is odd
            # and aa is a pth power mod p^2, then aa is a pth power mod any higher power of p

            if p % 2:

                # ZZ/(p^2)ZZ^\times is abstractly isomorphic to ZZ/(p)ZZ cross ZZ/(p-1)ZZ. then
                # aa is a pth power mod p^2 if (aa)^(p*(p-1)/p) == 1, ie if aa^(p-1) == 1.

                if Integers(p**2)(aa)**(p - 1) != 1:
                    return False

            # Otherwise, p=2.  By the strong statement of Hensel's lemma, if aa is a pth power
            # mod p^3, then it is a pth power mod higher powers of p.  So we need only check if it
            # is a pth power mod p^2 and p^3.

            elif ee == 2:

                # all odd squares a 1 mod 4

                if aa % 4 != 1:
                    return False

            # all odd squares are 1 mod 8

            elif aa % 8 != 1:
                return False

    return True


def _estimated_time(M2, M1, length, p):
    """
    Find the estimated time to extend congruences mod M1 to consistent congruences mod M2.

    INPUT:

    - ``M2`` -- integer; (the new modulus)

    - ``M1`` -- integer; (the old modulus)

    - ``length`` -- list (the current length of the list of congruences mod ``M1``)

    - ``p`` -- a prime

    OUTPUT: the estimated run time of the "CRT" step to combine consistent congruences

    EXAMPLES::

        sage: from sage.combinat.binary_recurrence_sequences import _estimated_time
        sage: _estimated_time(2**4*3**2*5*7*11*13*17, 2**4*3**2*5*7*11*13, 20, 7)       # needs sage.symbolic
        106.211159309421
    """
    # The heuristic run time of the CRT step to go from modulus M1 to M2

    # length is the current length of cong

    Q = p * log(M2)  # Size of our primes.
    NPrimes = log(M2 / M1) / log(Q)  # The number of primes

    return (length * (Q / p)**NPrimes).n()


# Find the list of necessary congruences for the index n of binary recurrence
# sequence R using the fact that the reduction mod ell must be a pth power
def _find_cong1(p, R, ell):
    """
    Find the list of permissible indices `n` for which `u_n = y^p` mod ``ell``.

    INPUT:

    - ``p`` -- a prime number

    - ``R`` -- an object in class :class:`BinaryRecurrenceSequence`

    - ``ell`` -- a prime number

    OUTPUT:

    - A list of permissible values of `n` modulo ``period(ell)`` and the integer ``period(ell)``.

    EXAMPLES::

        sage: R = BinaryRecurrenceSequence(1,1)
        sage: sage.combinat.binary_recurrence_sequences._find_cong1(7, R, 29)           # needs sage.rings.finite_rings
        ([0, 1, 2, 12, 13], 14)
    """
    F = GF(ell)
    u0 = F(R.u0)
    u1 = F(R.u1)
    bf, cf = F(R.b), F(R.c)
    a0 = u0
    a1 = u1      # a0 and a1 are variables for terms in sequence

    # The set of pth powers mod ell
    PPowers = set(i**p for i in F)

    # The period of R mod ell
    modu = R.period(ell)

    # cong1 keeps track of congruences mod modu for the sequence mod ell
    cong1 = []

    for n in range(modu):  # n is the index of the a0

        # Check whether a0 is a perfect power mod ell
        if a0 in PPowers:
            # if a0 is a perfect power mod ell, add the index
            # to the list of necessary congruences
            cong1.append(n)

        a0, a1 = a1, bf * a1 + cf * a0      # step up the variables

    cong1.sort()

    return cong1, modu


def _is_p_power(a, p) -> bool:
    """
    Determine whether `a` is a perfect `p`-th power.

    INPUT:

    - ``a`` -- integer

    - ``p`` -- a prime number

    OUTPUT: boolean

    EXAMPLES::

        sage: from sage.combinat.binary_recurrence_sequences import _is_p_power
        sage: _is_p_power(2**7, 7)
        True
        sage: _is_p_power(2**7*3**2, 7)
        False
    """
    try:
        Integer(a).nth_root(p)
    except ValueError:
        return False
    return True
